#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Jitter Explorer
===============

Application to explore orientation angle and angular dispersity.

From the command line::

    # Show docs
    python -m sasmodels.jitter --help

    # Guyou projection jitter, uniform over 20 degree theta and 10 in phi
    python -m sasmodels.jitter --projection=guyou --dist=uniform --jitter=20,10,0

From a jupyter cell::

    import ipyvolume as ipv
    from sasmodels import jitter
    import importlib; importlib.reload(jitter)
    jitter.set_plotter("ipv")

    size = (10, 40, 100)
    view = (20, 0, 0)

    #size = (15, 15, 100)
    #view = (60, 60, 0)

    dview = (0, 0, 0)
    #dview = (5, 5, 0)
    #dview = (15, 180, 0)
    #dview = (180, 15, 0)

    projection = 'equirectangular'
    #projection = 'azimuthal_equidistance'
    #projection = 'guyou'
    #projection = 'sinusoidal'
    #projection = 'azimuthal_equal_area'

    dist = 'uniform'
    #dist = 'gaussian'

    jitter.run(size=size, view=view, jitter=dview, dist=dist, projection=projection)
    #filename = projection+('_theta' if dview[0] == 180 else '_phi' if dview[1] == 180 else '')
    #ipv.savefig(filename+'.png')
"""
from __future__ import division, print_function

import argparse

import numpy as np
from numpy import pi, cos, sin, sqrt, exp, log, degrees, radians, arccos, arctan2

# Too many complaints about variable names from pylint:
#    a, b, c, u, v, x, y, z, dx, dy, dz, px, py, pz, R, Rx, Ry, Rz, ...
# pylint: disable=invalid-name

def draw_beam(axes, view=(0, 0), alpha=0.5, steps=25):
    """
    Draw the beam going from source at (0, 0, 1) to detector at (0, 0, -1)
    """
    #axes.plot([0,0],[0,0],[1,-1])
    #axes.scatter([0]*100,[0]*100,np.linspace(1, -1, 100), alpha=alpha)

    u = np.linspace(0, 2 * pi, steps)
    v = np.linspace(-1, 1, 2)

    r = 0.02
    x = r*np.outer(cos(u), np.ones_like(v))
    y = r*np.outer(sin(u), np.ones_like(v))
    z = 1.3*np.outer(np.ones_like(u), v)

    theta, phi = view
    shape = x.shape
    points = np.array([x.flatten(), y.flatten(), z.flatten()])
    points = Rz(phi)@Ry(theta)@points
    x, y, z = [v.reshape(shape) for v in points]
    axes.plot_surface(x, y, z, color='yellow', alpha=alpha)

    # TODO: draw endcaps on beam
    ## Drawing tiny balls on the end will work
    #draw_sphere(axes, radius=0.02, center=(0, 0, 1.3), color='yellow', alpha=alpha)
    #draw_sphere(axes, radius=0.02, center=(0, 0, -1.3), color='yellow', alpha=alpha)
    ## The following does not work
    #triangles = [(0, i+1, i+2) for i in range(steps-2)]
    #x_cap, y_cap = x[:, 0], y[:, 0]
    #for z_cap in z[:, 0], z[:, -1]:
    #    axes.plot_trisurf(x_cap, y_cap, z_cap, triangles,
    #                      color='yellow', alpha=alpha)


def draw_ellipsoid(axes, size, view, jitter, steps=25, alpha=1):
    """Draw an ellipsoid."""
    a, b, c = size
    u = np.linspace(0, 2 * pi, steps)
    v = np.linspace(0, pi, steps)
    x = a*np.outer(cos(u), sin(v))
    y = b*np.outer(sin(u), sin(v))
    z = c*np.outer(np.ones_like(u), cos(v))
    x, y, z = transform_xyz(view, jitter, x, y, z)

    axes.plot_surface(x, y, z, color='w', alpha=alpha)

    draw_labels(axes, view, jitter, [
        ('c+', [+0, +0, +c], [+1, +0, +0]),
        ('c-', [+0, +0, -c], [+0, +0, -1]),
        ('a+', [+a, +0, +0], [+0, +0, +1]),
        ('a-', [-a, +0, +0], [+0, +0, -1]),
        ('b+', [+0, +b, +0], [-1, +0, +0]),
        ('b-', [+0, -b, +0], [-1, +0, +0]),
    ])

def draw_sc(axes, size, view, jitter, steps=None, alpha=1):
    """Draw points for simple cubic paracrystal"""
    # pylint: disable=unused-argument
    atoms = _build_sc()
    _draw_crystal(axes, size, view, jitter, atoms=atoms)

def draw_fcc(axes, size, view, jitter, steps=None, alpha=1):
    """Draw points for face-centered cubic paracrystal"""
    # pylint: disable=unused-argument
    # Build the simple cubic crystal
    atoms = _build_sc()
    # Define the centers for each face
    # x planes at -1, 0, 1 have four centers per plane, at +/- 0.5 in y and z
    x, y, z = (
        [-1]*4 + [0]*4 + [1]*4,
        ([-0.5]*2 + [0.5]*2)*3,
        [-0.5, 0.5]*12,
    )
    # y and z planes can be generated by substituting x for y and z respectively
    atoms.extend(zip(x+y+z, y+z+x, z+x+y))
    _draw_crystal(axes, size, view, jitter, atoms=atoms)

def draw_bcc(axes, size, view, jitter, steps=None, alpha=1):
    """Draw points for body-centered cubic paracrystal"""
    # pylint: disable=unused-argument
    # Build the simple cubic crystal
    atoms = _build_sc()
    # Define the centers for each octant
    # x plane at +/- 0.5 have four centers per plane at +/- 0.5 in y and z
    x, y, z = (
        [-0.5]*4 + [0.5]*4,
        ([-0.5]*2 + [0.5]*2)*2,
        [-0.5, 0.5]*8,
    )
    atoms.extend(zip(x, y, z))
    _draw_crystal(axes, size, view, jitter, atoms=atoms)

def _draw_crystal(axes, size, view, jitter, atoms=None):
    atoms, size = np.asarray(atoms, 'd').T, np.asarray(size, 'd')
    x, y, z = atoms*size[:, None]
    x, y, z = transform_xyz(view, jitter, x, y, z)
    axes.scatter([x[0]], [y[0]], [z[0]], c='yellow', marker='o')
    axes.scatter(x[1:], y[1:], z[1:], c='r', marker='o')

def _build_sc():
    # three planes of 9 dots for x at -1, 0 and 1
    x, y, z = (
        [-1]*9 + [0]*9 + [1]*9,
        ([-1]*3 + [0]*3 + [1]*3)*3,
        [-1, 0, 1]*9,
    )
    atoms = list(zip(x, y, z))
    #print(list(enumerate(atoms)))
    # Pull the dot at (0, 0, 1) to the front of the list
    # It will be highlighted in the view
    index = 14
    highlight = atoms[index]
    del atoms[index]
    atoms.insert(0, highlight)
    return atoms

def draw_box(axes, size, view):
    """Draw a wireframe box at a particular view."""
    a, b, c = size
    x = a*np.array([+1, -1, +1, -1, +1, -1, +1, -1])
    y = b*np.array([+1, +1, -1, -1, +1, +1, -1, -1])
    z = c*np.array([+1, +1, +1, +1, -1, -1, -1, -1])
    x, y, z = transform_xyz(view, None, x, y, z)
    def _draw(i, j):
        axes.plot([x[i], x[j]], [y[i], y[j]], [z[i], z[j]], color='black')
    _draw(0, 1)
    _draw(0, 2)
    _draw(0, 3)
    _draw(7, 4)
    _draw(7, 5)
    _draw(7, 6)

def draw_parallelepiped(axes, size, view, jitter, steps=None,
                        color=(0.6, 1.0, 0.6), alpha=1):
    """Draw a parallelepiped surface, with view and jitter."""
    # pylint: disable=unused-argument
    a, b, c = size
    x = a*np.array([+1, -1, +1, -1, +1, -1, +1, -1])
    y = b*np.array([+1, +1, -1, -1, +1, +1, -1, -1])
    z = c*np.array([+1, +1, +1, +1, -1, -1, -1, -1])
    tri = np.array([
        # counter clockwise triangles
        # z: up/down, x: right/left, y: front/back
        [0, 1, 2], [3, 2, 1], # top face
        [6, 5, 4], [5, 6, 7], # bottom face
        [0, 2, 6], [6, 4, 0], # right face
        [1, 5, 7], [7, 3, 1], # left face
        [2, 3, 6], [7, 6, 3], # front face
        [4, 1, 0], [5, 1, 4], # back face
    ])

    x, y, z = transform_xyz(view, jitter, x, y, z)
    axes.plot_trisurf(x, y, triangles=tri, Z=z, color=color, alpha=alpha,
                      linewidth=0)

    # Colour the c+ face of the box.
    # Since I can't control face color, instead draw a thin box situated just
    # in front of the "c+" face.  Use the c face so that rotations about psi
    # rotate that face.
    if 0: # pylint: disable=using-constant-test
        color = (1, 0.6, 0.6)  # pink
        x = a*np.array([+1, -1, +1, -1, +1, -1, +1, -1])
        y = b*np.array([+1, +1, -1, -1, +1, +1, -1, -1])
        z = c*np.array([+1, +1, +1, +1, -1, -1, -1, -1])
        x, y, z = transform_xyz(view, jitter, x, y, abs(z)+0.001)
        axes.plot_trisurf(x, y, triangles=tri, Z=z, color=color, alpha=alpha)

    draw_labels(axes, view, jitter, [
        ('c+', [+0, +0, +c], [+1, +0, +0]),
        ('c-', [+0, +0, -c], [+0, +0, -1]),
        ('a+', [+a, +0, +0], [+0, +0, +1]),
        ('a-', [-a, +0, +0], [+0, +0, -1]),
        ('b+', [+0, +b, +0], [-1, +0, +0]),
        ('b-', [+0, -b, +0], [-1, +0, +0]),
    ])

def draw_sphere(axes, radius=1.0, steps=25,
                center=(0, 0, 0), color='w', alpha=1.):
    """Draw a sphere"""
    u = np.linspace(0, 2 * pi, steps)
    v = np.linspace(0, pi, steps)

    x = radius * np.outer(cos(u), sin(v)) + center[0]
    y = radius * np.outer(sin(u), sin(v)) + center[1]
    z = radius * np.outer(np.ones(np.size(u)), cos(v)) + center[2]
    axes.plot_surface(x, y, z, color=color, alpha=alpha)
    #axes.plot_wireframe(x, y, z)

def draw_axes(axes, origin=(-1, -1, -1), length=(2, 2, 2)):
    """Draw wireframe axes lines, with given origin and length"""
    x, y, z = origin
    dx, dy, dz = length
    axes.plot([x, x+dx], [y, y], [z, z], color='black')
    axes.plot([x, x], [y, y+dy], [z, z], color='black')
    axes.plot([x, x], [y, y], [z, z+dz], color='black')

def draw_person_on_sphere(axes, view, height=0.5, radius=1.0):
    """
    Draw a person on the surface of a sphere.

    *view* indicates (latitude, longitude, orientation)
    """
    limb_offset = height * 0.05
    head_radius = height * 0.10
    head_height = height - head_radius
    neck_length = head_radius * 0.50
    shoulder_height = height - 2*head_radius - neck_length
    torso_length = shoulder_height * 0.55
    torso_radius = torso_length * 0.30
    leg_length = shoulder_height - torso_length
    arm_length = torso_length * 0.90

    def _draw_part(y, z):
        x = np.zeros_like(y)
        xp, yp, zp = transform_xyz(view, None, x, y, z + radius)
        axes.plot(xp, yp, zp, color='k')

    # circle for head
    u = np.linspace(0, 2 * pi, 40)
    y = head_radius * cos(u)
    z = head_radius * sin(u) + head_height
    _draw_part(y, z)

    # rectangle for body
    y = np.array([-torso_radius, torso_radius, torso_radius, -torso_radius, -torso_radius])
    z = np.array([0., 0, torso_length, torso_length, 0]) + leg_length
    _draw_part(y, z)

    # arms
    y = np.array([-torso_radius - limb_offset, -torso_radius - limb_offset, -torso_radius])
    z = np.array([shoulder_height - arm_length, shoulder_height, shoulder_height])
    _draw_part(y, z)
    _draw_part(-y, z)  # pylint: disable=invalid-unary-operand-type

    # legs
    y = np.array([-torso_radius + limb_offset, -torso_radius + limb_offset])
    z = np.array([0, leg_length])
    _draw_part(y, z)
    _draw_part(-y, z)  # pylint: disable=invalid-unary-operand-type

    limits = [-radius-height, radius+height]
    axes.set_xlim(limits)
    axes.set_ylim(limits)
    axes.set_zlim(limits)
    axes.set_axis_off()

def draw_jitter(axes, view, jitter, dist='gaussian',
                size=(0.1, 0.4, 1.0),
                draw_shape=draw_parallelepiped,
                projection='equirectangular',
                alpha=0.8,
                views=None):
    """
    Represent jitter as a set of shapes at different orientations.
    """
    project, project_weight = get_projection(projection)

    # set max diagonal to 0.95
    scale = 0.95/sqrt(sum(v**2 for v in size))
    size = tuple(scale*v for v in size)

    dtheta, dphi, dpsi = jitter
    base = {'gaussian':3, 'rectangle':sqrt(3), 'uniform':1}[dist]
    def _steps(delta):
        if views is None:
            n = max(3, min(25, 2*int(base*delta/5)))
        else:
            n = views
        return base*delta*np.linspace(-1, 1, n) if delta > 0 else [0.]
    for theta in _steps(dtheta):
        for phi in _steps(dphi):
            for psi in _steps(dpsi):
                w = project_weight(theta, phi, 1.0, 1.0)
                if w > 0:
                    dview = project(theta, phi, psi)
                    draw_shape(axes, size, view, dview, alpha=alpha)
    for v in 'xyz':
        a, b, c = size
        lim = sqrt(a**2 + b**2 + c**2)
        getattr(axes, 'set_'+v+'lim')([-lim, lim])
        #getattr(axes, v+'axis').label.set_text(v)

PROJECTIONS = [
    # in order of PROJECTION number; do not change without updating the
    # constants in kernel_iq.c
    'equirectangular', 'sinusoidal', 'guyou', 'azimuthal_equidistance',
    'azimuthal_equal_area',
]
def get_projection(projection):

    """
    jitter projections
    <https://en.wikipedia.org/wiki/List_of_map_projections>

    equirectangular (standard latitude-longitude mesh)
        <https://en.wikipedia.org/wiki/Equirectangular_projection>
        Allows free movement in phi (around the equator), but theta is
        limited to +/- 90, and points are cos-weighted. Jitter in phi is
        uniform in weight along a line of latitude.  With small theta and
        phi ranging over +/- 180 this forms a wobbling disk.  With small
        phi and theta ranging over +/- 90 this forms a wedge like a slice
        of an orange.
    azimuthal_equidistance (Postel)
        <https://en.wikipedia.org/wiki/Azimuthal_equidistant_projection>
        Preserves distance from center, and so is an excellent map for
        representing a bivariate gaussian on the surface.  Theta and phi
        operate identically, cutting wegdes from the antipode of the viewing
        angle.  This unfortunately does not allow free movement in either
        theta or phi since the orthogonal wobble decreases to 0 as the body
        rotates through 180 degrees.
    sinusoidal (Sanson-Flamsteed, Mercator equal-area)
        <https://en.wikipedia.org/wiki/Sinusoidal_projection>
        Preserves arc length with latitude, giving bad behaviour at
        theta near +/- 90.  Theta and phi operate somewhat differently,
        so a system with a-b-c dtheta-dphi-dpsi will not give the same
        value as one with b-a-c dphi-dtheta-dpsi, as would be the case
        for azimuthal equidistance.  Free movement using theta or phi
        uniform over +/- 180 will work, but not as well as equirectangular
        phi, with theta being slightly worse.  Computationally it is much
        cheaper for wide theta-phi meshes since it excludes points which
        lie outside the sinusoid near theta +/- 90 rather than packing
        them close together as in equirectangle.  Note that the poles
        will be slightly overweighted for theta > 90 with the circle
        from theta at 90+dt winding backwards around the pole, overlapping
        the circle from theta at 90-dt.
    Guyou (hemisphere-in-a-square) **not weighted**
        <https://en.wikipedia.org/wiki/Guyou_hemisphere-in-a-square_projection>
        With tiling, allows rotation in phi or theta through +/- 180, with
        uniform spacing.  Both theta and phi allow free rotation, with wobble
        in the orthogonal direction reasonably well behaved (though not as
        good as equirectangular phi). The forward/reverse transformations
        relies on elliptic integrals that are somewhat expensive, so the
        behaviour has to be very good to justify the cost and complexity.
        The weighting function for each point has not yet been computed.
        Note: run the module *guyou.py* directly and it will show the forward
        and reverse mappings.
    azimuthal_equal_area  **incomplete**
        <https://en.wikipedia.org/wiki/Lambert_azimuthal_equal-area_projection>
        Preserves the relative density of the surface patches.  Not that
        useful and not completely implemented
    Gauss-Kreuger **not implemented**
        <https://en.wikipedia.org/wiki/Transverse_Mercator_projection#Ellipsoidal_transverse_Mercator>
        Should allow free movement in theta, but phi is distorted.
    """
    # pylint: disable=unused-argument
    # TODO: try Kent distribution instead of a gaussian warped by projection

    if projection == 'equirectangular':  #define PROJECTION 1
        def _project(theta_i, phi_j, psi):
            latitude, longitude = theta_i, phi_j
            return latitude, longitude, psi, 'xyz'
            #return Rx(phi_j)@Ry(theta_i)
        def _weight(theta_i, phi_j, w_i, w_j):
            return w_i*w_j*abs(cos(radians(theta_i)))
    elif projection == 'sinusoidal':  #define PROJECTION 2
        def _project(theta_i, phi_j, psi):
            latitude = theta_i
            scale = cos(radians(latitude))
            longitude = phi_j/scale if abs(phi_j) < abs(scale)*180 else 0
            #print("(%+7.2f, %+7.2f) => (%+7.2f, %+7.2f)"%(theta_i, phi_j, latitude, longitude))
            return latitude, longitude, psi, 'xyz'
            #return Rx(longitude)@Ry(latitude)
        def _project(theta_i, phi_j, w_i, w_j):
            latitude = theta_i
            scale = cos(radians(latitude))
            active = 1 if abs(phi_j) < abs(scale)*180 else 0
            return active*w_i*w_j
    elif projection == 'guyou':  #define PROJECTION 3  (eventually?)
        def _project(theta_i, phi_j, psi):
            from .guyou import guyou_invert
            #latitude, longitude = guyou_invert([theta_i], [phi_j])
            longitude, latitude = guyou_invert([phi_j], [theta_i])
            return latitude, longitude, psi, 'xyz'
            #return Rx(longitude[0])@Ry(latitude[0])
        def _weight(theta_i, phi_j, w_i, w_j):
            return w_i*w_j
    elif projection == 'azimuthal_equidistance':
        # Note that calculates angles for Rz Ry rather than Rx Ry
        def _project(theta_i, phi_j, psi):
            latitude = sqrt(theta_i**2 + phi_j**2)
            longitude = degrees(arctan2(phi_j, theta_i))
            #print("(%+7.2f, %+7.2f) => (%+7.2f, %+7.2f)"%(theta_i, phi_j, latitude, longitude))
            return latitude, longitude, psi-longitude, 'zyz'
            #R = Rz(longitude)@Ry(latitude)@Rz(psi)
            #return R_to_xyz(R)
            #return Rz(longitude)@Ry(latitude)
        def _weight(theta_i, phi_j, w_i, w_j):
            # Weighting for each point comes from the integral:
            #     \int\int I(q, lat, log) sin(lat) dlat dlog
            # We are doing a conformal mapping from disk to sphere, so we need
            # a change of variables g(theta, phi) -> (lat, long):
            #     lat, long = sqrt(theta^2 + phi^2), arctan(phi/theta)
            # giving:
            #     dtheta dphi = det(J) dlat dlong
            # where J is the jacobian from the partials of g. Using
            #     R = sqrt(theta^2 + phi^2),
            # then
            #     J = [[x/R, Y/R], -y/R^2, x/R^2]]
            # and
            #     det(J) = 1/R
            # with the final integral being:
            #    \int\int I(q, theta, phi) sin(R)/R dtheta dphi
            #
            # This does approximately the right thing, decreasing the weight
            # of each point as you go farther out on the disk, but it hasn't
            # yet been checked against the 1D integral results. Prior
            # to declaring this "good enough" and checking that integrals
            # work in practice, we will examine alternative mappings.
            #
            # The issue is that the mapping does not support the case of free
            # rotation about a single axis correctly, with a small deviation
            # in the orthogonal axis independent of the first axis.  Like the
            # usual polar coordiates integration, the integrated sections
            # form wedges, though at least in this case the wedge cuts through
            # the entire sphere, and treats theta and phi identically.
            latitude = sqrt(theta_i**2 + phi_j**2)
            weight = sin(radians(latitude))/latitude if latitude != 0 else 1
            return weight*w_i*w_j if latitude < 180 else 0
    elif projection == 'azimuthal_equal_area':
        # Note that calculates angles for Rz Ry rather than Rx Ry
        def _project(theta_i, phi_j, psi):
            radius = min(1, sqrt(theta_i**2 + phi_j**2)/180)
            latitude = 180-degrees(2*arccos(radius))
            longitude = degrees(arctan2(phi_j, theta_i))
            #print("(%+7.2f, %+7.2f) => (%+7.2f, %+7.2f)"%(theta_i, phi_j, latitude, longitude))
            return latitude, longitude, psi, 'zyz'
            #R = Rz(longitude)@Ry(latitude)@Rz(psi)
            #return R_to_xyz(R)
            #return Rz(longitude)@Ry(latitude)
        def _weight(theta_i, phi_j, w_i, w_j):
            latitude = sqrt(theta_i**2 + phi_j**2)
            weight = sin(radians(latitude))/latitude if latitude != 0 else 1
            return weight*w_i*w_j if latitude < 180 else 0
    else:
        raise ValueError("unknown projection %r"%projection)

    return _project, _weight

def R_to_xyz(R):
    """
    Return phi, theta, psi Tait-Bryan angles corresponding to the given rotation matrix.

    Extracting Euler Angles from a Rotation Matrix
    Mike Day, Insomniac Games
    https://d3cw3dd2w32x2b.cloudfront.net/wp-content/uploads/2012/07/euler-angles1.pdf
    Based on: Shoemake’s "Euler Angle Conversion", Graphics Gems IV, pp.  222-229
    """
    phi = arctan2(R[1, 2], R[2, 2])
    theta = arctan2(-R[0, 2], sqrt(R[0, 0]**2 + R[0, 1]**2))
    psi = arctan2(R[0, 1], R[0, 0])
    return degrees(phi), degrees(theta), degrees(psi)

def draw_mesh(axes, view, jitter, radius=1.2, n=11, dist='gaussian',
              projection='equirectangular'):
    """
    Draw the dispersion mesh showing the theta-phi orientations at which
    the model will be evaluated.
    """

    _project, _weight = get_projection(projection)
    def _rotate(theta, phi, z):
        dview = _project(theta, phi, 0.)
        if dview[3] == 'zyz':
            return Rz(dview[1])@Ry(dview[0])@z
        else:  # dview[3] == 'xyz':
            return Rx(dview[1])@Ry(dview[0])@z


    dist_x = np.linspace(-1, 1, n)
    weights = np.ones_like(dist_x)
    if dist == 'gaussian':
        dist_x *= 3
        weights = exp(-0.5*dist_x**2)
    elif dist == 'rectangle':
        # Note: uses sasmodels ridiculous definition of rectangle width
        dist_x *= sqrt(3)
    elif dist == 'uniform':
        pass
    else:
        raise ValueError("expected dist to be gaussian, rectangle or uniform")

    # mesh in theta, phi formed by rotating z
    dtheta, dphi, dpsi = jitter  # pylint: disable=unused-variable
    z = np.array([[0], [0], [radius]])
    points = np.hstack([_rotate(theta_i, phi_j, z)
                        for theta_i in dtheta*dist_x
                        for phi_j in dphi*dist_x])
    dist_w = np.array([_weight(theta_i, phi_j, w_i, w_j)
                       for w_i, theta_i in zip(weights, dtheta*dist_x)
                       for w_j, phi_j in zip(weights, dphi*dist_x)])
    #print(max(dist_w), min(dist_w), min(dist_w[dist_w > 0]))
    points = points[:, dist_w > 0]
    dist_w = dist_w[dist_w > 0]
    dist_w /= max(dist_w)

    # rotate relative to beam
    points = orient_relative_to_beam(view, points)

    x, y, z = [np.array(v).flatten() for v in points]
    #plt.figure(2); plt.clf(); plt.hist(z, bins=np.linspace(-1, 1, 51))
    axes.scatter(x, y, z, c=dist_w, marker='o', vmin=0., vmax=1.)

def draw_labels(axes, view, jitter, text):
    """
    Draw text at a particular location.
    """
    labels, locations, orientations = zip(*text)
    px, py, pz = zip(*locations)
    dx, dy, dz = zip(*orientations)

    px, py, pz = transform_xyz(view, jitter, px, py, pz)
    dx, dy, dz = transform_xyz(view, jitter, dx, dy, dz)

    # TODO: zdir for labels is broken, and labels aren't appearing.
    for label, p, zdir in zip(labels, zip(px, py, pz), zip(dx, dy, dz)):
        zdir = np.asarray(zdir).flatten()
        axes.text(p[0], p[1], p[2], label, zdir=zdir)

# Definition of rotation matrices comes from wikipedia:
#    https://en.wikipedia.org/wiki/Rotation_matrix#Basic_rotations
def Rx(angle):
    """Construct a matrix to rotate points about *x* by *angle* degrees."""
    angle = radians(angle)
    rot = [[1, 0, 0],
           [0, +cos(angle), -sin(angle)],
           [0, +sin(angle), +cos(angle)]]
    return np.matrix(rot)

def Ry(angle):
    """Construct a matrix to rotate points about *y* by *angle* degrees."""
    angle = radians(angle)
    rot = [[+cos(angle), 0, +sin(angle)],
           [0, 1, 0],
           [-sin(angle), 0, +cos(angle)]]
    return np.matrix(rot)

def Rz(angle):
    """Construct a matrix to rotate points about *z* by *angle* degrees."""
    angle = radians(angle)
    rot = [[+cos(angle), -sin(angle), 0],
           [+sin(angle), +cos(angle), 0],
           [0, 0, 1]]
    return np.matrix(rot)

def transform_xyz(view, jitter, x, y, z):
    """
    Send a set of (x,y,z) points through the jitter and view transforms.
    """
    x, y, z = [np.asarray(v) for v in (x, y, z)]
    shape = x.shape
    points = np.matrix([x.flatten(), y.flatten(), z.flatten()])
    points = apply_jitter(jitter, points)
    points = orient_relative_to_beam(view, points)
    x, y, z = [np.array(v).reshape(shape) for v in points]
    return x, y, z

def apply_jitter(jitter, points):
    """
    Apply the jitter transform to a set of points.

    Points are stored in a 3 x n numpy matrix, not a numpy array or tuple.
    """
    if jitter is None:
        return points
    # Hack to deal with the fact that azimuthal_equidistance uses euler angles
    if len(jitter) == 4:
        dtheta, dphi, dpsi, _ = jitter
        points = Rz(dphi)@Ry(dtheta)@Rz(dpsi)@points
    else:
        dtheta, dphi, dpsi = jitter
        points = Rx(dphi)@Ry(dtheta)@Rz(dpsi)@points
    return points

def orient_relative_to_beam(view, points):
    """
    Apply the view transform to a set of points.

    Points are stored in a 3 x n numpy matrix, not a numpy array or tuple.
    """
    theta, phi, psi = view
    points = Rz(phi)@Ry(theta)@Rz(psi)@points # viewing angle
    #points = Rz(psi)@Ry(pi/2-theta)@Rz(phi)@points # 1-D integration angles
    #points = Rx(phi)@Ry(theta)@Rz(psi)@points  # angular dispersion angle
    return points

def orient_relative_to_beam_quaternion(view, points):
    """
    Apply the view transform to a set of points.

    Points are stored in a 3 x n numpy matrix, not a numpy array or tuple.

    This variant uses quaternions rather than rotation matrices for the
    computation.  It works but it is not used because it doesn't solve
    any problems.  The challenge of mapping theta/phi/psi to SO(3) does
    not disappear by calculating the transform differently.
    """
    theta, phi, psi = view
    x, y, z = [1, 0, 0], [0, 1, 0], [0, 0, 1]
    q = Quaternion(1, [0, 0, 0])
    ## Compose a rotation about the three axes by rotating
    ## the unit vectors before applying the rotation.
    #q = Quaternion.from_angle_axis(theta, q.rot(x)) * q
    #q = Quaternion.from_angle_axis(phi, q.rot(y)) * q
    #q = Quaternion.from_angle_axis(psi, q.rot(z)) * q
    ## The above turns out to be equivalent to reversing
    ## the order of application, so ignore it and use below.
    q = q * Quaternion.from_angle_axis(theta, x)
    q = q * Quaternion.from_angle_axis(phi, y)
    q = q * Quaternion.from_angle_axis(psi, z)
    ## Reverse the order by post-multiply rather than pre-multiply
    #q = Quaternion.from_angle_axis(theta, x) * q
    #q = Quaternion.from_angle_axis(phi, y) * q
    #q = Quaternion.from_angle_axis(psi, z) * q
    #print("axes psi", q.rot(np.matrix([x, y, z]).T))
    return q.rot(points)
#orient_relative_to_beam = orient_relative_to_beam_quaternion

# === Quaterion class definition === BEGIN
# Simple stand-alone quaternion class

# Note: this code works but isn't unused since quaternions didn't solve the
# representation problem.  Leave it here in case we want to revisit this later.

#import numpy as np
class Quaternion(object):
    r"""
    Quaternion(w, r) = w + ir[0] + jr[1] + kr[2]

    Quaternion.from_angle_axis(theta, r) for a rotation of angle theta about
    an axis oriented toward the direction r.  This defines a unit quaternion,
    normalizing $r$ to the unit vector $\hat r$, and setting quaternion
    $Q = \cos \theta + \sin \theta \hat r$

    Quaternion objects can be multiplied, which applies a rotation about the
    given axis, allowing composition of rotations without risk of gimbal lock.
    The resulting quaternion is applied to a set of points using *Q.rot(v)*.
    """
    def __init__(self, w, r):
        self.w = w
        self.r = np.asarray(r, dtype='d')

    @staticmethod
    def from_angle_axis(theta, r):
        """Build quaternion as rotation theta about axis r"""
        theta = np.radians(theta)/2
        r = np.asarray(r)
        w = np.cos(theta)
        r = np.sin(theta)*r/np.dot(r, r)
        return Quaternion(w, r)

    def __mul__(self, other):
        """Multiply quaterions"""
        if isinstance(other, Quaternion):
            w = self.w*other.w - np.dot(self.r, other.r)
            r = self.w*other.r + other.w*self.r + np.cross(self.r, other.r)
            return Quaternion(w, r)
        raise NotImplementedError("Quaternion * non-quaternion not implemented")

    def rot(self, v):
        """Transform point *v* by quaternion"""
        v = np.asarray(v).T
        use_transpose = (v.shape[-1] != 3)
        if use_transpose:
            v = v.T
        v = v + np.cross(2*self.r, np.cross(self.r, v) + self.w*v)
        #v = v + 2*self.w*np.cross(self.r, v) + np.cross(2*self.r, np.cross(self.r, v))
        if use_transpose:
            v = v.T
        return v.T

    def conj(self):
        """Conjugate quaternion"""
        return Quaternion(self.w, -self.r)

    def inv(self):
        """Inverse quaternion"""
        return self.conj()/self.norm()**2

    def norm(self):
        """Quaternion length"""
        return np.sqrt(self.w**2 + np.sum(self.r**2))

    def __str__(self):
        return "%g%+gi%+gj%+gk"%(self.w, self.r[0], self.r[1], self.r[2])

def test_qrot():
    """Quaternion checks"""
    # Define rotation of 60 degrees around an axis in y-z that is 60 degrees
    # from y.  The rotation axis is determined by rotating the point [0, 1, 0]
    # about x.
    ax = Quaternion.from_angle_axis(60, [1, 0, 0]).rot([0, 1, 0])
    q = Quaternion.from_angle_axis(60, ax)
    # Set the point to be rotated, and its expected rotated position.
    p = [1, -1, 2]
    target = [(10+4*np.sqrt(3))/8, (1+2*np.sqrt(3))/8, (14-3*np.sqrt(3))/8]
    #print(q, q.rot(p) - target)
    assert max(abs(q.rot(p) - target)) < 1e-14
#test_qrot()
#import sys; sys.exit()
# === Quaterion class definition === END

# translate between number of dimension of dispersity and the number of
# points along each dimension.
PD_N_TABLE = {
    (0, 0, 0): (0, 0, 0),     # 0
    (1, 0, 0): (100, 0, 0),   # 100
    (0, 1, 0): (0, 100, 0),
    (0, 0, 1): (0, 0, 100),
    (1, 1, 0): (30, 30, 0),   # 900
    (1, 0, 1): (30, 0, 30),
    (0, 1, 1): (0, 30, 30),
    (1, 1, 1): (15, 15, 15),  # 3375
}

def clipped_range(data, portion=1.0, mode='central'):
    """
    Determine range from data.

    If *portion* is 1, use full range, otherwise use the center of the range
    or the top of the range, depending on whether *mode* is 'central' or 'top'.
    """
    if portion < 1.0:
        if mode == 'central':
            data = np.sort(data.flatten())
            offset = int(portion*len(data)/2 + 0.5)
            return data[offset], data[-offset]
        if mode == 'top':
            data = np.sort(data.flatten())
            offset = int(portion*len(data) + 0.5)
            return data[offset], data[-1]
    # Default: full range
    return data.min(), data.max()

def draw_scattering(calculator, axes, view, jitter, dist='gaussian'):
    """
    Plot the scattering for the particular view.

    *calculator* is returned from :func:`build_model`.  *axes* are the 3D axes
    on which the data will be plotted.  *view* and *jitter* are the current
    orientation and orientation dispersity.  *dist* is one of the sasmodels
    weight distributions.
    """
    if dist == 'uniform':  # uniform is not yet in this branch
        dist, scale = 'rectangle', 1/sqrt(3)
    else:
        scale = 1

    # add the orientation parameters to the model parameters
    theta, phi, psi = view
    theta_pd, phi_pd, psi_pd = [scale*v for v in jitter]
    theta_pd_n, phi_pd_n, psi_pd_n = PD_N_TABLE[(theta_pd > 0, phi_pd > 0, psi_pd > 0)]
    ## increase pd_n for testing jitter integration rather than simple viz
    #theta_pd_n, phi_pd_n, psi_pd_n = [5*v for v in (theta_pd_n, phi_pd_n, psi_pd_n)]

    pars = dict(
        theta=theta, theta_pd=theta_pd, theta_pd_type=dist, theta_pd_n=theta_pd_n,
        phi=phi, phi_pd=phi_pd, phi_pd_type=dist, phi_pd_n=phi_pd_n,
        psi=psi, psi_pd=psi_pd, psi_pd_type=dist, psi_pd_n=psi_pd_n,
    )
    pars.update(calculator.pars)

    # compute the pattern
    qx, qy = calculator.qxqy
    Iqxy = calculator(**pars).reshape(len(qx), len(qy))

    # scale it and draw it
    Iqxy = log(Iqxy)
    if calculator.limits:
        # use limits from orientation (0,0,0)
        vmin, vmax = calculator.limits
    else:
        vmax = Iqxy.max()
        vmin = vmax*10**-7
        #vmin, vmax = clipped_range(Iqxy, portion=portion, mode='top')
    #vmin, vmax = Iqxy.min(), Iqxy.max()
    #print("range",(vmin,vmax))
    #qx, qy = np.meshgrid(qx, qy)
    if 0:  # pylint: disable=using-constant-test
        level = np.asarray(255*(Iqxy - vmin)/(vmax - vmin), 'i')
        level[level < 0] = 0
        from matplotlib import pylab as plt
        colors = plt.get_cmap()(level)
        #from matplotlib import cm
        #colors = cm.coolwarm(level)
        #colors = cm.gist_yarg(level)
        #colors = cm.Wistia(level)
        colors[level <= 0, 3] = 0.  # set floor to transparent
        x, y = np.meshgrid(qx/qx.max(), qy/qy.max())
        axes.plot_surface(x, y, -1.1*np.ones_like(x), facecolors=colors)
    elif 1:  # pylint: disable=using-constant-test
        axes.contourf(qx/qx.max(), qy/qy.max(), Iqxy, zdir='z', offset=-1.1,
                      levels=np.linspace(vmin, vmax, 24))
    else:
        axes.pcolormesh(qx, qy, Iqxy)

def build_model(model_name, n=150, qmax=0.5, **pars):
    """
    Build a calculator for the given shape.

    *model_name* is any sasmodels model.  *n* and *qmax* define an n x n mesh
    on which to evaluate the model.  The remaining parameters are stored in
    the returned calculator as *calculator.pars*.  They are used by
    :func:`draw_scattering` to set the non-orientation parameters in the
    calculation.

    Returns a *calculator* function which takes a dictionary or parameters and
    produces Iqxy.  The Iqxy value needs to be reshaped to an n x n matrix
    for plotting.  See the :class:`.direct_model.DirectModel` class
    for details.
    """
    from sasmodels.core import load_model_info, build_model as build_sasmodel
    from sasmodels.data import empty_data2D
    from sasmodels.direct_model import DirectModel

    model_info = load_model_info(model_name)
    model = build_sasmodel(model_info) #, dtype='double!')
    q = np.linspace(-qmax, qmax, n)
    data = empty_data2D(q, q)
    calculator = DirectModel(data, model)

    # Remember the data axes so we can plot the results
    calculator.qxqy = (q, q)

    # stuff the values for non-orientation parameters into the calculator
    calculator.pars = pars.copy()
    calculator.pars.setdefault('background', 1e-3)

    # fix the data limits so that we can see if the pattern fades
    # under rotation or angular dispersion
    Iqxy = calculator(theta=0, phi=0, psi=0, **calculator.pars)
    Iqxy = log(Iqxy)
    vmin, vmax = clipped_range(Iqxy, 0.95, mode='top')
    calculator.limits = vmin, vmax+1

    return calculator

def select_calculator(model_name, n=150, size=(10, 40, 100)):
    """
    Create a model calculator for the given shape.

    *model_name* is one of sphere, cylinder, ellipsoid, triaxial_ellipsoid,
    parallelepiped or bcc_paracrystal. *n* is the number of points to use
    in the q range.  *qmax* is chosen based on model parameters for the
    given model to show something intersting.

    Returns *calculator* and tuple *size* (a,b,c) giving minor and major
    equitorial axes and polar axis respectively.  See :func:`build_model`
    for details on the returned calculator.
    """
    a, b, c = size
    d_factor = 0.06  # for paracrystal models
    if model_name == 'sphere':
        calculator = build_model(
            'sphere', n=n, radius=c)
        a = b = c
    elif model_name == 'sc_paracrystal':
        a = b = c
        dnn = c
        radius = 0.5*c
        calculator = build_model(
            'sc_paracrystal', n=n,
            dnn=dnn, d_factor=d_factor, radius=(1-d_factor)*radius,
            background=0)
    elif model_name == 'fcc_paracrystal':
        a = b = c
        # nearest neigbour distance dnn should be 2 radius, but I think the
        # model uses lattice spacing rather than dnn in its calculations
        dnn = 0.5*c
        radius = sqrt(2)/4 * c
        calculator = build_model(
            'fcc_paracrystal', n=n,
            dnn=dnn, d_factor=d_factor, radius=(1-d_factor)*radius,
            background=0)
    elif model_name == 'bcc_paracrystal':
        a = b = c
        # nearest neigbour distance dnn should be 2 radius, but I think the
        # model uses lattice spacing rather than dnn in its calculations
        dnn = 0.5*c
        radius = sqrt(3)/2 * c
        calculator = build_model(
            'bcc_paracrystal', n=n,
            dnn=dnn, d_factor=d_factor, radius=(1-d_factor)*radius,
            background=0)
    elif model_name == 'cylinder':
        calculator = build_model(
            'cylinder', n=n, qmax=0.3, radius=b, length=c)
        a = b
    elif model_name == 'ellipsoid':
        calculator = build_model(
            'ellipsoid', n=n, qmax=1.0,
            radius_polar=c, radius_equatorial=b)
        a = b
    elif model_name == 'triaxial_ellipsoid':
        calculator = build_model(
            'triaxial_ellipsoid', n=n, qmax=0.5,
            radius_equat_minor=a, radius_equat_major=b, radius_polar=c)
    elif model_name == 'parallelepiped':
        calculator = build_model(
            'parallelepiped', n=n, length_a=a, length_b=b, length_c=c)
    else:
        raise ValueError("unknown model %s"%model_name)

    return calculator, (a, b, c)

SHAPES = [
    'parallelepiped',
    'sphere', 'ellipsoid', 'triaxial_ellipsoid',
    'cylinder',
    'fcc_paracrystal', 'bcc_paracrystal', 'sc_paracrystal',
]

DRAW_SHAPES = {
    'fcc_paracrystal': draw_fcc,
    'bcc_paracrystal': draw_bcc,
    'sc_paracrystal': draw_sc,
    'parallelepiped': draw_parallelepiped,
}

DISTRIBUTIONS = [
    'gaussian', 'rectangle', 'uniform', 'cyclic',
]
DIST_LIMITS = {
    'gaussian': 30,
    'rectangle': 90/sqrt(3),
    'uniform': 90,
}


def run(model_name='parallelepiped', size=(10, 40, 100),
        view=(0, 0, 0), jitter=(0, 0, 0),
        dist='gaussian', mesh=30,
        projection='equirectangular'):
    """
    Show an interactive orientation and jitter demo.

    *model_name* is one of: sphere, ellipsoid, triaxial_ellipsoid,
    parallelepiped, cylinder, or sc/fcc/bcc_paracrystal

    *size* gives the dimensions (a, b, c) of the shape.

    *view* gives the initial view (theta, phi, psi) of the shape.

    *view* gives the initial jitter (dtheta, dphi, dpsi) of the shape.

    *dist* is the type of dispersition: gaussian, rectangle, or uniform.

    *mesh* is the number of points in the dispersion mesh.

    *projection* is the map projection to use for the mesh: equirectangular,
    sinusoidal, guyou, azimuthal_equidistance, or azimuthal_equal_area.
    """
    # projection number according to 1-order position in list, but
    # only 1 and 2 are implemented so far.
    from sasmodels import generate
    generate.PROJECTION = PROJECTIONS.index(projection) + 1
    if generate.PROJECTION > 2:
        print("*** PROJECTION %s not implemented in scattering function ***"%projection)
        generate.PROJECTION = 2

    # set up calculator
    calculator, size = select_calculator(model_name, n=150, size=size)
    draw_shape = DRAW_SHAPES.get(model_name, draw_parallelepiped)
    #draw_shape = draw_fcc

    ## uncomment to set an independent the colour range for every view
    ## If left commented, the colour range is fixed for all views
    calculator.limits = None

    PLOT_ENGINE(calculator, draw_shape, size, view, jitter, dist, mesh, projection)

def _mpl_plot(calculator, draw_shape, size, view, jitter, dist, mesh, projection):
    # Note: travis-ci does not support mpl_toolkits.mplot3d, but this shouldn't be
    # an issue since we are lazy-loading the package on a path that isn't tested.
    # Importing mplot3d adds projection='3d' option to subplot
    import mpl_toolkits.mplot3d  # pylint: disable=unused-import
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from matplotlib.widgets import Slider

    ## create the plot window
    #plt.hold(True)
    plt.subplots(num=None, figsize=(5.5, 5.5))
    plt.set_cmap('gist_earth')
    plt.clf()
    plt.gcf().canvas.manager.set_window_title(projection)
    #gs = gridspec.GridSpec(2,1,height_ratios=[4,1])
    #axes = plt.subplot(gs[0], projection='3d')
    axes = plt.axes([0.0, 0.2, 1.0, 0.8], projection='3d')
    try:  # CRUFT: not all versions of matplotlib accept 'square' 3d projection
        axes.axis('square')
    except Exception:
        pass

    # CRUFT: use axisbg instead of facecolor for matplotlib<2
    facecolor_prop = 'facecolor' if mpl.__version__ > '2' else 'axisbg'
    props = {facecolor_prop: 'lightgoldenrodyellow'}

    ## add control widgets to plot
    axes_theta = plt.axes([0.05, 0.15, 0.50, 0.04], **props)
    axes_phi = plt.axes([0.05, 0.10, 0.50, 0.04], **props)
    axes_psi = plt.axes([0.05, 0.05, 0.50, 0.04], **props)
    stheta = Slider(axes_theta, u'θ', -90, 90, valinit=0)
    sphi = Slider(axes_phi, u'φ', -180, 180, valinit=0)
    spsi = Slider(axes_psi, u'ψ', -180, 180, valinit=0)

    axes_dtheta = plt.axes([0.70, 0.15, 0.20, 0.04], **props)
    axes_dphi = plt.axes([0.70, 0.1, 0.20, 0.04], **props)
    axes_dpsi = plt.axes([0.70, 0.05, 0.20, 0.04], **props)

    # Note: using ridiculous definition of rectangle distribution, whose width
    # in sasmodels is sqrt(3) times the given width.  Divide by sqrt(3) to keep
    # the maximum width to 90.
    dlimit = DIST_LIMITS[dist]
    sdtheta = Slider(axes_dtheta, u'Δθ', 0, 2*dlimit, valinit=0)
    sdphi = Slider(axes_dphi, u'Δφ', 0, 2*dlimit, valinit=0)
    sdpsi = Slider(axes_dpsi, u'Δψ', 0, 2*dlimit, valinit=0)

    ## initial view and jitter
    theta, phi, psi = view
    stheta.set_val(theta)
    sphi.set_val(phi)
    spsi.set_val(psi)
    dtheta, dphi, dpsi = jitter
    sdtheta.set_val(dtheta)
    sdphi.set_val(dphi)
    sdpsi.set_val(dpsi)

    ## callback to draw the new view
    def _update(val, axis=None):
        # pylint: disable=unused-argument
        view = stheta.val, sphi.val, spsi.val
        jitter = sdtheta.val, sdphi.val, sdpsi.val
        # set small jitter as 0 if multiple pd dims
        dims = sum(v > 0 for v in jitter)
        limit = [0, 0.5, 5, 5][dims]
        jitter = [0 if v < limit else v for v in jitter]
        axes.cla()

        ## Visualize as person on globe
        #draw_sphere(axes, radius=0.5)
        #draw_person_on_sphere(axes, view, radius=0.5)

        ## Move beam instead of shape
        #draw_beam(axes, -view[:2])
        #draw_jitter(axes, (0,0,0), (0,0,0), views=3)

        ## Move shape and draw scattering
        draw_beam(axes, (0, 0), alpha=1.)
        #draw_person_on_sphere(axes, view, radius=1.2, height=0.5)
        draw_jitter(axes, view, jitter, dist=dist, size=size, alpha=1.,
                    draw_shape=draw_shape, projection=projection, views=3)
        draw_mesh(axes, view, jitter, dist=dist, n=mesh, projection=projection)
        draw_scattering(calculator, axes, view, jitter, dist=dist)

        plt.gcf().canvas.draw()

    ## bind control widgets to view updater
    stheta.on_changed(lambda v: _update(v, 'theta'))
    sphi.on_changed(lambda v: _update(v, 'phi'))
    spsi.on_changed(lambda v: _update(v, 'psi'))
    sdtheta.on_changed(lambda v: _update(v, 'dtheta'))
    sdphi.on_changed(lambda v: _update(v, 'dphi'))
    sdpsi.on_changed(lambda v: _update(v, 'dpsi'))

    ## initialize view
    _update(None, 'phi')

    ## go interactive
    plt.show()


def map_colors(z, kw):
    """
    Process matplotlib-style colour arguments.

    Pulls 'cmap', 'alpha', 'vmin', and 'vmax' from th *kw* dictionary, setting
    the *kw['color']* to an RGB array.  These are ignored if 'c' or 'color' are
    set inside *kw*.
    """
    from matplotlib import cm

    cmap = kw.pop('cmap', cm.coolwarm)
    alpha = kw.pop('alpha', None)
    vmin = kw.pop('vmin', z.min())
    vmax = kw.pop('vmax', z.max())
    c = kw.pop('c', None)
    color = kw.pop('color', c)
    if color is None:
        znorm = ((z - vmin) / (vmax - vmin)).clip(0, 1)
        color = cmap(znorm)
    elif isinstance(color, np.ndarray) and color.shape == z.shape:
        color = cmap(color)
    if alpha is None:
        if isinstance(color, np.ndarray):
            color = color[..., :3]
    else:
        color[..., 3] = alpha
    kw['color'] = color

def make_vec(*args):
    """Turn all elements of *args* into numpy arrays"""
    #return [np.asarray(v, 'd').flatten() for v in args]
    return [np.asarray(v, 'd') for v in args]

def make_image(z, kw):
    """Convert numpy array *z* into a *PIL* RGB image."""
    import PIL.Image
    from matplotlib import cm

    cmap = kw.pop('cmap', cm.coolwarm)

    znorm = (z-z.min())/z.ptp()
    c = cmap(znorm)
    c = c[..., :3]
    rgb = np.asarray(c*255, 'u1')
    image = PIL.Image.fromarray(rgb, mode='RGB')
    return image


_IPV_MARKERS = {
    'o': 'sphere',
}
_IPV_COLORS = {
    'w': 'white',
    'k': 'black',
    'c': 'cyan',
    'm': 'magenta',
    'y': 'yellow',
    'r': 'red',
    'g': 'green',
    'b': 'blue',
}
def _ipv_fix_color(kw):
    alpha = kw.pop('alpha', None)
    color = kw.get('color', None)
    if isinstance(color, str):
        color = _IPV_COLORS.get(color, color)
        kw['color'] = color
    if alpha is not None:
        color = kw['color']
        #TODO: convert color to [r, g, b, a] if not already
        if isinstance(color, (tuple, list)):
            if len(color) == 3:
                color = (color[0], color[1], color[2], alpha)
            else:
                color = (color[0], color[1], color[2], alpha*color[3])
            color = np.array(color)
        if isinstance(color, np.ndarray) and color.shape[-1] == 4:
            color[..., 3] = alpha
            kw['color'] = color

def _ipv_set_transparency(kw, obj):
    color = kw.get('color', None)
    if (isinstance(color, np.ndarray)
            and color.shape[-1] == 4
            and (color[..., 3] != 1.0).any()):
        obj.material.transparent = True
        obj.material.side = "FrontSide"

def ipv_axes():
    """
    Build a matplotlib style Axes interface for ipyvolume
    """
    import ipyvolume as ipv

    class Axes(object):
        """
        Matplotlib Axes3D style interface to ipyvolume renderer.
        """
        # pylint: disable=no-self-use,no-init
        # transparency can be achieved by setting the following:
        #    mesh.color = [r, g, b, alpha]
        #    mesh.material.transparent = True
        #    mesh.material.side = "FrontSide"
        # smooth(ish) rotation can be achieved by setting:
        #    slide.continuous_update = True
        #    figure.animation = 0.
        #    mesh.material.x = x
        #    mesh.material.y = y
        #    mesh.material.z = z
        # maybe need to synchronize update of x/y/z to avoid shimmy when moving
        def plot(self, x, y, z, **kw):
            """mpl style plot interface for ipyvolume"""
            _ipv_fix_color(kw)
            x, y, z = make_vec(x, y, z)
            ipv.plot(x, y, z, **kw)
        def plot_surface(self, x, y, z, **kw):
            """mpl style plot_surface interface for ipyvolume"""
            facecolors = kw.pop('facecolors', None)
            if facecolors is not None:
                kw['color'] = facecolors
            _ipv_fix_color(kw)
            x, y, z = make_vec(x, y, z)
            h = ipv.plot_surface(x, y, z, **kw)
            _ipv_set_transparency(kw, h)
            #h.material.side = "DoubleSide"
            return h
        def plot_trisurf(self, x, y, triangles=None, Z=None, **kw):
            """mpl style plot_trisurf interface for ipyvolume"""
            kw.pop('linewidth', None)
            _ipv_fix_color(kw)
            x, y, z = make_vec(x, y, Z)
            if triangles is not None:
                triangles = np.asarray(triangles)
            h = ipv.plot_trisurf(x, y, z, triangles=triangles, **kw)
            _ipv_set_transparency(kw, h)
            return h
        def scatter(self, x, y, z, **kw):
            """mpl style scatter interface for ipyvolume"""
            x, y, z = make_vec(x, y, z)
            map_colors(z, kw)
            marker = kw.get('marker', None)
            kw['marker'] = _IPV_MARKERS.get(marker, marker)
            h = ipv.scatter(x, y, z, **kw)
            _ipv_set_transparency(kw, h)
            return h
        def contourf(self, x, y, v, zdir='z', offset=0, levels=None, **kw):
            """mpl style contourf interface for ipyvolume"""
            # pylint: disable=unused-argument
            # Don't use contour for now (although we might want to later)
            self.pcolor(x, y, v, zdir='z', offset=offset, **kw)
        def pcolor(self, x, y, v, zdir='z', offset=0, **kw):
            """mpl style pcolor interface for ipyvolume"""
            # pylint: disable=unused-argument
            x, y, v = make_vec(x, y, v)
            image = make_image(v, kw)
            xmin, xmax = x.min(), x.max()
            ymin, ymax = y.min(), y.max()
            x = np.array([[xmin, xmax], [xmin, xmax]])
            y = np.array([[ymin, ymin], [ymax, ymax]])
            z = x*0 + offset
            u = np.array([[0., 1], [0, 1]])
            v = np.array([[0., 0], [1, 1]])
            h = ipv.plot_mesh(x, y, z, u=u, v=v, texture=image, wireframe=False)
            _ipv_set_transparency(kw, h)
            h.material.side = "DoubleSide"
            return h
        def text(self, *args, **kw):
            """mpl style text interface for ipyvolume"""
            pass
        def set_xlim(self, limits):
            """mpl style set_xlim interface for ipyvolume"""
            ipv.xlim(*limits)
        def set_ylim(self, limits):
            """mpl style set_ylim interface for ipyvolume"""
            ipv.ylim(*limits)
        def set_zlim(self, limits):
            """mpl style set_zlim interface for ipyvolume"""
            ipv.zlim(*limits)
        def set_axes_on(self):
            """mpl style set_axes_on interface for ipyvolume"""
            ipv.style.axes_on()
        def set_axis_off(self):
            """mpl style set_axes_off interface for ipyvolume"""
            ipv.style.axes_off()
    return Axes()

def _ipv_plot(calculator, draw_shape, size, view, jitter, dist, mesh, projection):
    from IPython.display import display
    import ipywidgets as widgets
    import ipyvolume as ipv

    axes = ipv_axes()

    def _draw(view, jitter):
        camera = ipv.gcf().camera
        #print(ipv.gcf().__dict__.keys())
        #print(dir(ipv.gcf()))
        ipv.figure(animation=0.)  # no animation when updating object mesh

        # set small jitter as 0 if multiple pd dims
        dims = sum(v > 0 for v in jitter)
        limit = [0, 0.5, 5, 5][dims]
        jitter = [0 if v < limit else v for v in jitter]

        ## Visualize as person on globe
        #draw_beam(axes, (0, 0))
        #draw_sphere(axes, radius=0.5)
        #draw_person_on_sphere(axes, view, radius=0.5)

        ## Move beam instead of shape
        #draw_beam(axes, view=(-view[0], -view[1]))
        #draw_jitter(axes, view=(0,0,0), jitter=(0,0,0))

        ## Move shape and draw scattering
        draw_beam(axes, (0, 0), steps=25)
        draw_jitter(axes, view, jitter, dist=dist, size=size, alpha=1.0,
                    draw_shape=draw_shape, projection=projection)
        draw_mesh(axes, view, jitter, dist=dist, n=mesh, radius=0.95,
                  projection=projection)
        draw_scattering(calculator, axes, view, jitter, dist=dist)

        draw_axes(axes, origin=(-1, -1, -1.1))
        ipv.style.box_off()
        ipv.style.axes_off()
        ipv.xyzlabel(" ", " ", " ")

        ipv.gcf().camera = camera
        ipv.show()


    trange, prange = (-180., 180., 1.), (-180., 180., 1.)
    dtrange, dprange = (0., 180., 1.), (0., 180., 1.)

    ## Super simple interfaca, but uses non-ascii variable namese
    # θ φ ψ Δθ Δφ Δψ
    #def update(**kw):
    #    view = kw['θ'], kw['φ'], kw['ψ']
    #    jitter = kw['Δθ'], kw['Δφ'], kw['Δψ']
    #    draw(view, jitter)
    #widgets.interact(update, θ=trange, φ=prange, ψ=prange, Δθ=dtrange, Δφ=dprange, Δψ=dprange)

    def _update(theta, phi, psi, dtheta, dphi, dpsi):
        _draw(view=(theta, phi, psi), jitter=(dtheta, dphi, dpsi))

    def _slider(name, steps, init=0.):
        return widgets.FloatSlider(
            value=init,
            min=steps[0],
            max=steps[1],
            step=steps[2],
            description=name,
            disabled=False,
            #continuous_update=True,
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='.1f',
            )
    theta = _slider(u'θ', trange, view[0])
    phi = _slider(u'φ', prange, view[1])
    psi = _slider(u'ψ', prange, view[2])
    dtheta = _slider(u'Δθ', dtrange, jitter[0])
    dphi = _slider(u'Δφ', dprange, jitter[1])
    dpsi = _slider(u'Δψ', dprange, jitter[2])
    fields = {
        'theta': theta, 'phi': phi, 'psi': psi,
        'dtheta': dtheta, 'dphi': dphi, 'dpsi': dpsi,
    }
    ui = widgets.HBox([
        widgets.VBox([theta, phi, psi]),
        widgets.VBox([dtheta, dphi, dpsi])
    ])

    out = widgets.interactive_output(_update, fields)
    display(ui, out)


_ENGINES = {
    "matplotlib": _mpl_plot,
    "mpl": _mpl_plot,
    #"plotly": _plotly_plot,
    "ipvolume": _ipv_plot,
    "ipv": _ipv_plot,
}
PLOT_ENGINE = _ENGINES["matplotlib"]
def set_plotter(name):
    """
    Setting the plotting engine to matplotlib/ipyvolume or equivalently mpl/ipv.
    """
    global PLOT_ENGINE
    PLOT_ENGINE = _ENGINES[name]

def main():
    """
    Command line interface to the jitter viewer.
    """
    parser = argparse.ArgumentParser(
        description="Display jitter",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
    parser.add_argument('-p', '--projection', choices=PROJECTIONS,
                        default=PROJECTIONS[0],
                        help='coordinate projection')
    parser.add_argument('-s', '--size', type=str, default='10,40,100',
                        help='a,b,c lengths')
    parser.add_argument('-v', '--view', type=str, default='0,0,0',
                        help='initial view angles')
    parser.add_argument('-j', '--jitter', type=str, default='0,0,0',
                        help='initial angular dispersion')
    parser.add_argument('-d', '--distribution', choices=DISTRIBUTIONS,
                        default=DISTRIBUTIONS[0],
                        help='jitter distribution')
    parser.add_argument('-m', '--mesh', type=int, default=30,
                        help='#points in theta-phi mesh')
    parser.add_argument('shape', choices=SHAPES, nargs='?', default=SHAPES[0],
                        help='oriented shape')
    opts = parser.parse_args()
    size = tuple(float(v) for v in opts.size.split(','))
    view = tuple(float(v) for v in opts.view.split(','))
    jitter = tuple(float(v) for v in opts.jitter.split(','))
    run(opts.shape, size=size, view=view, jitter=jitter,
        mesh=opts.mesh, dist=opts.distribution,
        projection=opts.projection)

if __name__ == "__main__":
    main()



import numpy as np
from numpy import exp, sin, cos, pi, radians, degrees

from sasmodels.weights import Dispersion as BaseDispersion

class Dispersion(BaseDispersion):
    r"""
    Cyclic gaussian dispersion on orientation.

    .. math:

        w(\theta) = e^{-\frac{\sin^2 \theta}{2 \sigma^2}}

    This provides a close match to the gaussian distribution for
    low angles, but the tails are limited to $\pm 90^\circ$.  For $\sigma$
    large the distribution is approximately uniform.  The usual polar coordinate
    projection applies, with $\theta$ weights scaled by $\cos \theta$
    and $\phi$ weights unscaled.

    This is eqivalent to a Maier-Saupe distribution with order
    parameter $a = 1/(2 \sigma^2)$, with $\sigma$ in radians.
    """
    type = "cyclic_gaussian"
    default = dict(npts=35, width=1, nsigmas=3)

    # Note: center is always zero for orientation distributions
    def _weights(self, center, sigma, lb, ub):
        # Convert sigma in degrees to radians
        sigma = radians(sigma)

        # Limit width to +/- 90 degrees
        width = min(self.nsigmas*sigma, pi/2)
        x = np.linspace(-width, width, self.npts)

        # Truncate the distribution in case the parameter value is limited
        x[(x >= radians(lb)) & (x <= radians(ub))]

        # Return orientation in degrees with Maier-Saupe weights
        return degrees(x), exp(-0.5*sin(x)**2/sigma**2)



"""
GPU driver for C kernels

TODO: docs are out of date

There should be a single GPU environment running on the system.  This
environment is constructed on the first call to :func:`environment`, and the
same environment is returned on each call.

After retrieving the environment, the next step is to create the kernel.
This is done with a call to :meth:`GpuEnvironment.compile_program`, which
returns the type of data used by the kernel.

Next a :class:`GpuInput` object should be created with the correct kind
of data.  This data object can be used by multiple kernels, for example,
if the target model is a weighted sum of multiple kernels.  The data
should include any extra evaluation points required to compute the proper
data smearing.  This need not match the square grid for 2D data if there
is an index saying which q points are active.

Together the GpuInput, the program, and a device form a :class:`GpuKernel`.
This kernel is used during fitting, receiving new sets of parameters and
evaluating them.  The output value is stored in an output buffer on the
devices, where it can be combined with other structure factors and form
factors and have instrumental resolution effects applied.

In order to use OpenCL for your models, you will need OpenCL drivers for
your machine.  These should be available from your graphics card vendor.
Intel provides OpenCL drivers for CPUs as well as their integrated HD
graphics chipsets.  AMD also provides drivers for Intel CPUs, but as of
this writing the performance is lacking compared to the Intel drivers.
NVidia combines drivers for CUDA and OpenCL in one package.  The result
is a bit messy if you have multiple drivers installed.  You can see which
drivers are available by starting python and running:

    import pyopencl as cl
    cl.create_some_context(interactive=True)

Once you have done that, it will show the available drivers which you
can select.  It will then tell you that you can use these drivers
automatically by setting the SAS_OPENCL environment variable, which is
PYOPENCL_CTX equivalent but not conflicting with other pyopnecl programs.

Some graphics cards have multiple devices on the same card.  You cannot
yet use both of them concurrently to evaluate models, but you can run
the program twice using a different device for each session.

OpenCL kernels are compiled when needed by the device driver.  Some
drivers produce compiler output even when there is no error.  You
can see the output by setting PYOPENCL_COMPILER_OUTPUT=1.  It should be
harmless, albeit annoying.
"""
from __future__ import print_function

import sys
import os
import warnings
import logging
import time

try:
    from time import perf_counter as clock
except ImportError: # CRUFT: python < 3.3
    if sys.platform.count("darwin") > 0:
        from time import time as clock
    else:
        from time import clock

import numpy as np  # type: ignore

# Attempt to setup OpenCL. This may fail if the pyopencl package is not
# installed or if it is installed but there are no devices available.
try:
    import pyopencl as cl  # type: ignore
    from pyopencl import mem_flags as mf
    from pyopencl.characterize import get_fast_inaccurate_build_options
    # CRUFT: pyopencl<2019.1.2 is breaking on intel drivers for mac
    from pyopencl.version import VERSION_TEXT
    if sys.platform == 'darwin' and VERSION_TEXT < '2019.1.2':
        os.environ['PYOPENCL_NO_CACHE'] = '1'
    # Ask OpenCL for the default context so that we know that one exists.
    cl.create_some_context(interactive=False)
    HAVE_OPENCL = True
    OPENCL_ERROR = ""
except Exception as exc:
    HAVE_OPENCL = False
    OPENCL_ERROR = str(exc)

from . import generate
from .generate import F32, F64
from .kernel import KernelModel, Kernel

# pylint: disable=unused-import
try:
    from typing import Tuple, Callable, Any, List, Dict
    from .modelinfo import ModelInfo
    from .details import CallDetails
except ImportError:
    pass
# pylint: enable=unused-import


# CRUFT: pyopencl < 2017.1 (as of June 2016 needs quotes around include path).
def quote_path(v):
    # type: (str) -> str
    """
    Quote the path if it is not already quoted.

    If v starts with '-', then assume that it is a -I option or similar
    and do not quote it.  This is fragile:  -Ipath with space needs to
    be quoted.
    """
    return '"'+v+'"' if v and ' ' in v and not v[0] in "\"'-" else v


def fix_pyopencl_include():
    # type: () -> None
    """
    Monkey patch pyopencl to allow spaces in include file path.
    """
    # pylint: disable=protected-access
    import pyopencl
    if hasattr(pyopencl, '_DEFAULT_INCLUDE_OPTIONS'):
        pyopencl._DEFAULT_INCLUDE_OPTIONS = [
            quote_path(v) for v in pyopencl._DEFAULT_INCLUDE_OPTIONS
            ]


if HAVE_OPENCL:
    fix_pyopencl_include()

# The max loops number is limited by the amount of local memory available
# on the device.  You don't want to make this value too big because it will
# waste resources, nor too small because it may interfere with users trying
# to do their polydispersity calculations.  A value of 1024 should be much
# larger than necessary given that cost grows as npts^k where k is the number
# of polydisperse parameters.
MAX_LOOPS = 2048

# Pragmas for enable OpenCL features.  Be sure to protect them so that they
# still compile even if OpenCL is not present.
_F16_PRAGMA = """\
#if defined(__OPENCL_VERSION__) // && !defined(cl_khr_fp16)
#  pragma OPENCL EXTENSION cl_khr_fp16: enable
#endif
"""

_F64_PRAGMA = """\
#if defined(__OPENCL_VERSION__) // && !defined(cl_khr_fp64)
#  pragma OPENCL EXTENSION cl_khr_fp64: enable
#endif
"""


def use_opencl():
    # type: () -> bool
    """Return True if OpenCL is the default computational engine"""
    sas_opencl = os.environ.get("SAS_OPENCL", "OpenCL").lower()
    return HAVE_OPENCL and sas_opencl != "none" and not sas_opencl.startswith("cuda")


ENV = None
def reset_environment():
    # type: () -> "GpuEnvironment"
    """
    Return a new OpenCL context, such as after a change to SAS_OPENCL.
    """
    global ENV
    ENV = GpuEnvironment() if use_opencl() else None
    return ENV

def environment():
    # type: () -> "GpuEnvironment"
    """
    Returns a singleton :class:`GpuEnvironment`.

    This provides an OpenCL context and one queue per device.
    """
    if ENV is None:
        if not HAVE_OPENCL:
            raise RuntimeError("OpenCL startup failed with ***"
                               + OPENCL_ERROR + "***; using C compiler instead")
        reset_environment()
        if ENV is None:
            raise RuntimeError("SAS_OPENCL=None in environment")
    return ENV


def has_type(device, dtype):
    # type: (cl.Device, np.dtype) -> bool
    """
    Return true if device supports the requested precision.
    """
    if dtype == F32:
        return True
    elif dtype == F64:
        return "cl_khr_fp64" in device.extensions
    else:
        # Not supporting F16 type since it isn't accurate enough.
        return False


def get_warp(kernel, queue):
    # type: (cl.Kernel, cl.CommandQueue) -> int
    """
    Return the size of an execution batch for *kernel* running on *queue*.
    """
    return kernel.get_work_group_info(
        cl.kernel_work_group_info.PREFERRED_WORK_GROUP_SIZE_MULTIPLE,
        queue.device)


def compile_model(context, source, dtype, fast=False):
    # type: (cl.Context, str, np.dtype, bool) -> cl.Program
    """
    Build a model to run on the gpu.

    Returns the compiled program and its type.

    Raises an error if the desired precision is not available.
    """
    dtype = np.dtype(dtype)
    if not all(has_type(d, dtype) for d in context.devices):
        raise RuntimeError("%s not supported for devices"%dtype)

    source_list = [generate.convert_type(source, dtype)]

    if dtype == generate.F16:
        source_list.insert(0, _F16_PRAGMA)
    elif dtype == generate.F64:
        source_list.insert(0, _F64_PRAGMA)

    # Note: USE_SINCOS makes the Intel CPU slower under OpenCL.
    if context.devices[0].type == cl.device_type.GPU:
        source_list.insert(0, "#define USE_SINCOS\n")
    options = (get_fast_inaccurate_build_options(context.devices[0])
               if fast else [])
    source = "\n".join(source_list)
    program = cl.Program(context, source).build(options=options)

    #print("done with "+program)
    return program


# For now, this returns one device in the context.
# TODO: Create a context that contains all devices on all platforms.
class GpuEnvironment(object):
    """
    GPU context for OpenCL, with possibly many devices and one queue per device.
    """
    def __init__(self):
        # type: () -> None
        # Find gpu context.
        context_list = _create_some_context()

        # Find a context for F32 and for F64 (maybe the same one).
        # F16 isn't good enough.
        self.context = {}
        for dtype in (F32, F64):
            for context in context_list:
                if has_type(context.devices[0], dtype):
                    self.context[dtype] = context
                    break
            else:
                self.context[dtype] = None

        # Build a queue for each context.
        self.queue = {}
        context = self.context[F32]
        self.queue[F32] = cl.CommandQueue(context, context.devices[0])
        if self.context[F64] == self.context[F32]:
            self.queue[F64] = self.queue[F32]
        elif self.context[F64] is not None:
            context = self.context[F64]
            self.queue[F64] = cl.CommandQueue(context, context.devices[0])
        else:
            self.queue[F64] = None

        ## Byte boundary for data alignment.
        #self.data_boundary = max(context.devices[0].min_data_type_align_size
        #                         for context in self.context.values())

        # Cache for compiled programs, and for items in context.
        self.compiled = {}

    def has_type(self, dtype):
        # type: (np.dtype) -> bool
        """
        Return True if all devices support a given type.
        """
        return self.context.get(dtype, None) is not None

    def compile_program(self, name, source, dtype, fast, timestamp):
        # type: (str, str, np.dtype, bool, float) -> cl.Program
        """
        Compile the program for the device in the given context.
        """
        # Note: PyOpenCL caches based on md5 hash of source, options and device
        # but I'll do so as well just to save some data munging time.
        tag = generate.tag_source(source)
        key = "%s-%s-%s%s"%(name, dtype, tag, ("-fast" if fast else ""))
        # Check timestamp on program.
        program, program_timestamp = self.compiled.get(key, (None, np.inf))
        if program_timestamp < timestamp:
            del self.compiled[key]
        if key not in self.compiled:
            context = self.context[dtype]
            logging.info("building %s for OpenCL %s", key,
                         context.devices[0].name.strip())
            program = compile_model(self.context[dtype],
                                    str(source), dtype, fast)
            self.compiled[key] = (program, timestamp)
        return program


def _create_some_context():
    # type: () -> cl.Context
    """
    Protected call to cl.create_some_context without interactivity.

    Uses SAS_OPENCL or PYOPENCL_CTX if they are set in the environment,
    otherwise scans for the most appropriate device using
    :func:`_get_default_context`.  Ignore *SAS_OPENCL=OpenCL*, which
    indicates that an OpenCL device should be used without specifying
    which one (and not a CUDA device, or no GPU).
    """
    # Assume we do not get here if SAS_OPENCL is None or CUDA.
    sas_opencl = os.environ.get('SAS_OPENCL', 'opencl')
    if sas_opencl.lower() != 'opencl':
        # Setting PYOPENCL_CTX as a SAS_OPENCL to create cl context.
        os.environ["PYOPENCL_CTX"] = sas_opencl

    if 'PYOPENCL_CTX' in os.environ:
        try:
            return [cl.create_some_context(interactive=False)]
        except Exception as exc:
            # TODO: Should warnings instead be put into logging.warn?
            warnings.warn(str(exc))
            warnings.warn(
                "pyopencl.create_some_context() failed.  The environment "
                "variable 'SAS_OPENCL' or 'PYOPENCL_CTX' might not be set "
                "correctly")

    return _get_default_context()


def _get_default_context():
    # type: () -> List[cl.Context]
    """
    Get an OpenCL context, preferring GPU over CPU, and preferring Intel
    drivers over AMD drivers.
    """
    # Note: on mobile devices there is automatic clock scaling if either the
    # CPU or the GPU is underutilized; probably doesn't affect us, but we if
    # it did, it would mean that putting a busy loop on the CPU while the GPU
    # is running may increase throughput.
    #
    # MacBook Pro, base install:
    #     {'Apple': [Intel CPU, NVIDIA GPU]}
    # MacBook Pro, base install:
    #     {'Apple': [Intel CPU, Intel GPU]}
    # 2 x NVIDIA 295 with Intel and NVIDIA opencl drivers install:
    #     {'Intel': [CPU], 'NVIDIA': [GPU, GPU, GPU, GPU]}
    gpu, cpu = None, None
    for platform in cl.get_platforms():
        # AMD provides a much weaker CPU driver than Intel/Apple, so avoid it.
        # If someone has bothered to install the AMD/NVIDIA drivers, prefer
        # them over the integrated graphics driver that may have been supplied
        # with the CPU chipset.
        preferred_cpu = (platform.vendor.startswith('Intel')
                         or platform.vendor.startswith('Apple'))
        preferred_gpu = (platform.vendor.startswith('Advanced')
                         or platform.vendor.startswith('NVIDIA'))
        for device in platform.get_devices():
            if device.type == cl.device_type.GPU:
                # If the existing type is not GPU then it will be CUSTOM
                # or ACCELERATOR so don't override it.
                if gpu is None or (preferred_gpu and gpu.type == cl.device_type.GPU):
                    gpu = device
            elif device.type == cl.device_type.CPU:
                if cpu is None or preferred_cpu:
                    cpu = device
            else:
                # System has cl.device_type.ACCELERATOR or cl.device_type.CUSTOM
                # Intel Phi for example registers as an accelerator.
                # Since the user installed a custom device on their system
                # and went through the pain of sorting out OpenCL drivers for
                # it, lets assume they really do want to use it as their
                # primary compute device.
                gpu = device

    # Order the devices by gpu then by cpu; when searching for an available
    # device by data type they will be checked in this order, which means
    # that if the gpu supports double then the cpu will never be used (though
    # we may make it possible to explicitly request the cpu at some point).
    devices = []
    if gpu is not None:
        devices.append(gpu)
    if cpu is not None:
        devices.append(cpu)
    return [cl.Context([d]) for d in devices]


class GpuModel(KernelModel):
    """
    GPU wrapper for a single model.

    *source* and *model_info* are the model source and interface as returned
    from :func:`.generate.make_source` and :func:`.modelinfo.make_model_info`.

    *dtype* is the desired model precision.  Any numpy dtype for single
    or double precision floats will do, such as 'f', 'float32' or 'single'
    for single and 'd', 'float64' or 'double' for double.  Double precision
    is an optional extension which may not be available on all devices.
    Half precision ('float16','half') may be available on some devices.
    Fast precision ('fast') is a loose version of single precision, indicating
    that the compiler is allowed to take shortcuts.
    """
    info = None  # type: ModelInfo
    source = ""  # type: str
    dtype = None  # type: np.dtype
    fast = False  # type: bool
    _program = None  # type: cl.Program
    _kernels = None  # type: Dict[str, cl.Kernel]

    def __init__(self, source, model_info, dtype=generate.F32, fast=False):
        # type: (Dict[str,str], ModelInfo, np.dtype, bool) -> None
        #print("create model", id(self))
        self.info = model_info
        self.source = source
        self.dtype = dtype
        self.fast = fast
        # TODO: can a model be freed?

    def __getstate__(self):
        # type: () -> Tuple[ModelInfo, str, np.dtype, bool]
        return self.info, self.source, self.dtype, self.fast

    def __setstate__(self, state):
        # type: (Tuple[ModelInfo, str, np.dtype, bool]) -> None
        self.info, self.source, self.dtype, self.fast = state
        self._program = self._kernels = None

    def make_kernel(self, q_vectors):
        # type: (List[np.ndarray]) -> "GpuKernel"
        return GpuKernel(self, q_vectors)

    def get_function(self, name):
        # type: (str) -> cl.Kernel
        """
        Fetch the kernel from the environment by name, compiling it if it
        does not already exist.
        """
        if self._program is None:
            self._prepare_program()
        return self._kernels[name]

    def _prepare_program(self):
        # type: (str) -> None
        env = environment()
        timestamp = generate.ocl_timestamp(self.info)
        program = env.compile_program(
            self.info.name,
            self.source['opencl'],
            self.dtype,
            self.fast,
            timestamp)
        variants = ['Iq', 'Iqxy', 'Imagnetic']
        names = [generate.kernel_name(self.info, k) for k in variants]
        functions = [getattr(program, k) for k in names]
        self._kernels = {k: v for k, v in zip(variants, functions)}
        # Keep a handle to program so GC doesn't collect.
        self._program = program


# TODO: Check that we don't need a destructor for buffers which go out of scope.
class GpuInput(object):
    """
    Make q data available to the gpu.

    *q_vectors* is a list of q vectors, which will be *[q]* for 1-D data,
    and *[qx, qy]* for 2-D data.  Internally, the vectors will be reallocated
    to get the best performance on OpenCL, which may involve shifting and
    stretching the array to better match the memory architecture.  Additional
    points will be evaluated with *q=1e-3*.

    *dtype* is the data type for the q vectors. The data type should be
    set to match that of the kernel, which is an attribute of
    :class:`GpuModel`.  Note that not all kernels support double
    precision, so even if the program was created for double precision,
    the *GpuModel.dtype* may be single precision.

    Call :meth:`release` when complete.  Even if not called directly, the
    buffer will be released when the data object is freed.
    """
    nq = 0
    dtype = generate.F32
    is_2d = False
    q = None
    q_b = None
    def __init__(self, q_vectors, dtype=generate.F32):
        # type: (List[np.ndarray], np.dtype) -> None
        #print("create input", id(self))
        # TODO: Do we ever need double precision q?
        self.nq = q_vectors[0].size
        self.dtype = np.dtype(dtype)
        self.is_2d = (len(q_vectors) == 2)
        if self.nq == 0:
            raise ValueError("GpuInput vector length must be greater than zero")
        if self.is_2d and q_vectors[1].shape != q_vectors[0].shape:
            raise ValueError("GpuInput vectors must be the same shape")
        # TODO: Stretch input based on get_warp().
        # Not doing it now since warp depends on kernel, which is not known
        # at this point, so instead using 32, which is good on the set of
        # architectures tested so far.
        if self.is_2d:
            width = ((self.nq+15)//16)*16
            self.q = np.empty((width, 2), dtype=dtype)
            self.q[:self.nq, 0] = q_vectors[0]
            self.q[:self.nq, 1] = q_vectors[1]
        else:
            width = ((self.nq+31)//32)*32
            self.q = np.empty(width, dtype=dtype)
            self.q[:self.nq] = q_vectors[0]
        self.global_size = [self.q.shape[0]]
        #print("creating inputs of size", self.global_size)

        # Transfer input value to GPU.
        env = environment()
        context = env.context[self.dtype]
        self.q_b = cl.Buffer(context, mf.READ_ONLY | mf.COPY_HOST_PTR,
                             hostbuf=self.q)

    def release(self):
        # type: () -> None
        """
        Free the buffer associated with the q value.
        """
        #print("release input", id(self))
        if self.q_b is not None:
            self.q_b.release()
            self.q_b = None

    def __del__(self):
        # type: () -> None
        #print("input deleted", id(self))
        self.release()


class GpuKernel(Kernel):
    """
    Callable SAS kernel.

    *model* is the GpuModel object to call

    The kernel is derived from :class:`.kernel.Kernel`, providing the
    *_call_kernel()* method to evaluate the kernel for a given set of
    parameters.  Because of the need to move the q values to the GPU before
    evaluation, the kernel is instantiated for a particular set of q vectors,
    and can be called many times without transfering q each time.

    Call :meth:`release` when done with the kernel instance.
    """
    #: SAS model information structure.
    info = None  # type: ModelInfo
    #: Kernel precision.
    dtype = None  # type: np.dtype
    #: Kernel dimensions (1d or 2d).
    dim = ""  # type: str
    #: Calculation results, updated after each call to *_call_kernel()*.
    result = None  # type: np.ndarray
    q_input = None # type: GpuInput
    _result_b = None # type: cl.Buffer

    def __init__(self, model, q_vectors):
        # type: (GpuModel, List[np.ndarray]) -> None
        #print("create kernel", id(self))
        dtype = model.dtype
        self.q_input = GpuInput(q_vectors, dtype)
        self._model = model

        # Attributes accessed from the outside.
        self.dim = '2d' if self.q_input.is_2d else '1d'
        self.info = model.info
        self.dtype = dtype

        # Converter to translate input to target type.
        self._as_dtype = np.float64 if dtype == generate.F64 else np.float32

        # Holding place for the returned value.
        nout = 2 if self.info.have_Fq and self.dim == '1d' else 1
        extra_q = 4  # Total weight, form volume, shell volume and R_eff.
        self.result = np.empty(self.q_input.nq*nout + extra_q, dtype)

        # Allocate result value on GPU.
        env = environment()
        context = env.context[self.dtype]
        width = ((self.result.size+31)//32)*32 * self.dtype.itemsize
        self._result_b = cl.Buffer(context, mf.READ_WRITE, width)

    def _call_kernel(self, call_details, values, cutoff, magnetic,
                     radius_effective_mode):
        # type: (CallDetails, np.ndarray, float, bool, int) -> None
        env = environment()
        queue = env.queue[self._model.dtype]
        if queue is None:
            raise RuntimeError("No support for type %s in OpenCL"
                               % str(self._model.dtype))
        context = queue.context

        # Arrange data transfer to card.
        details_b = cl.Buffer(context, mf.READ_ONLY | mf.COPY_HOST_PTR,
                              hostbuf=call_details.buffer)
        values_b = cl.Buffer(context, mf.READ_ONLY | mf.COPY_HOST_PTR,
                             hostbuf=values)

        # Setup kernel function and arguments.
        name = 'Iq' if self.dim == '1d' else 'Imagnetic' if magnetic else 'Iqxy'
        kernel = self._model.get_function(name)
        kernel_args = [
            np.uint32(self.q_input.nq),  # Number of inputs.
            None,  # Placeholder for pd_start.
            None,  # Placeholder for pd_stop.
            details_b,  # Problem definition.
            values_b,  # Parameter values.
            self.q_input.q_b,  # Q values.
            self._result_b,   # Result storage.
            self._as_dtype(cutoff),  # Probability cutoff.
            np.uint32(radius_effective_mode),  # R_eff mode.
        ]

        # Call kernel and retrieve results.
        #print("Calling OpenCL")
        #call_details.show(values)
        wait_for = None
        last_nap = clock()
        step = 1000000//self.q_input.nq + 1
        for start in range(0, call_details.num_eval, step):
            stop = min(start + step, call_details.num_eval)
            #print("queuing",start,stop)
            kernel_args[1:3] = [np.int32(start), np.int32(stop)]
            wait_for = [kernel(queue, self.q_input.global_size, None,
                               *kernel_args, wait_for=wait_for)]
            if stop < call_details.num_eval:
                # Allow other processes to run.
                wait_for[0].wait()
                current_time = clock()
                if current_time - last_nap > 0.5:
                    time.sleep(0.001)
                    last_nap = current_time
        cl.enqueue_copy(queue, self.result, self._result_b, wait_for=wait_for)
        #print("result", self.result)

        # Free buffers.
        details_b.release()
        values_b.release()

    def release(self):
        # type: () -> None
        """
        Release resources associated with the kernel.
        """
        #print("release kernel", id(self))
        if self.q_input is not None:
            self.q_input.release()
            self.q_input = None
        if self._result_b is not None:
            self._result_b.release()
            self._result_b = None

    def __del__(self):
        # type: () -> None
        #print("delete kernel", id(self))
        self.release()



"""
Explore integration of rotationally symmetric shapes
"""

from __future__ import print_function, division

import os, sys
sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))

import numpy as np
from numpy import pi, sin, cos, sqrt, exp, expm1, degrees, log10
from numpy.polynomial.legendre import leggauss
from scipy.integrate import dblquad, simps, romb, romberg
import pylab

from sasmodels.special import square
from sasmodels.special import Gauss20Wt, Gauss20Z
from sasmodels.special import Gauss76Wt, Gauss76Z
from sasmodels.special import Gauss150Wt, Gauss150Z
from sasmodels.special import sas_2J1x_x, sas_sinx_x, sas_3j1x_x

SLD = 3.0
SLD_SOLVENT = 6
CONTRAST = SLD - SLD_SOLVENT

def make_cylinder(radius, length):
    def cylinder(qab, qc):
        return sas_2J1x_x(qab*radius) * sas_sinx_x(qc*0.5*length)
    cylinder.__doc__ = "cylinder radius=%g, length=%g"%(radius, length)
    volume = pi*radius**2*length
    norm = CONTRAST**2*volume/10000
    return norm, cylinder

def make_long_cylinder(radius, length):
    def long_cylinder(q):
        return norm/q * sas_2J1x_x(q*radius)**2
    long_cylinder.__doc__ = "long cylinder radius=%g, length=%g"%(radius, length)
    volume = pi*radius**2*length
    norm = CONTRAST**2*volume/10000*pi/length
    return long_cylinder

def make_sphere(radius):
    def sphere(qab, qc):
        q = sqrt(qab**2 + qc**2)
        return sas_3j1x_x(q*radius)
    sphere.__doc__ = "sphere radius=%g"%(radius,)
    volume = 4*pi*radius**3/3
    norm = CONTRAST**2*volume/10000
    return norm, sphere

THETA_LOW, THETA_HIGH = 0, pi/2
SCALE = 1


def kernel_1d(q, theta):
    """
    S(q) kernel for paracrystal forms.
    """
    qab = q*sin(theta)
    qc = q*cos(theta)
    return NORM*KERNEL(qab, qc)**2

def gauss_quad_1d(q, n=150):
    """
    Compute the integral using gaussian quadrature for n = 20, 76 or 150.
    """
    z, w = leggauss(n)
    theta = (THETA_HIGH-THETA_LOW)*(z + 1)/2 + THETA_LOW
    sin_theta = abs(sin(theta))
    Zq = kernel_1d(q=q, theta=theta)
    return np.sum(Zq*w*sin_theta)*(THETA_HIGH-THETA_LOW)/2

def gridded_1d(q, n=300):
    """
    Compute the integral on a regular grid using rectangular, trapezoidal,
    simpsons, and romberg integration.  Romberg integration requires that
    the grid be of size n = 2**k + 1.
    """
    theta = np.linspace(THETA_LOW, THETA_HIGH, n)
    Zq = kernel_1d(q=q, theta=theta)
    Zq *= abs(sin(theta))
    dx = theta[1]-theta[0]
    print("rect-%d"%n, np.sum(Zq)*dx*SCALE)
    print("trapz-%d"%n, np.trapz(Zq, dx=dx)*SCALE)
    print("simpson-%d"%n, simps(Zq, dx=dx)*SCALE)
    print("romb-%d"%n, romb(Zq, dx=dx)*SCALE)

def scipy_romberg_1d(q):
    """
    Compute the integral using romberg integration.  This function does not
    complete in a reasonable time.  No idea if it is accurate.
    """
    evals = [0]
    def outer(theta):
        evals[0] += 1
        return kernel_1d(q, theta=theta)*abs(sin(theta))
    result = romberg(outer, THETA_LOW, THETA_HIGH, divmax=100)*SCALE
    print("scipy romberg", evals[0], result)

def plot_1d(q, n=300):
    """
    Plot the function that needs to be integrated in order to compute
    the I(q) at a particular q.  *n* is the number of points in the grid.
    """
    theta = np.linspace(THETA_LOW, THETA_HIGH, n)
    Zq = kernel_1d(q=q, theta=theta)
    Zq *= abs(sin(theta))
    pylab.semilogy(degrees(theta), np.fmax(Zq, 1.e-6), label="Q=%g"%q)
    pylab.title("%s I(q, theta) sin(theta)" % (KERNEL.__doc__,))
    pylab.xlabel("theta (degrees)")
    pylab.ylabel("Iq 1/cm")

def Iq_trapz(q, n):
    theta = np.linspace(THETA_LOW, THETA_HIGH, n)
    Zq = kernel_1d(q=q, theta=theta)
    Zq *= abs(sin(theta))
    dx = theta[1]-theta[0]
    return np.trapz(Zq, dx=dx)*SCALE

def plot_Iq(q, n, form="trapz"):
    if form == "trapz":
        Iq = np.array([Iq_trapz(qk, n) for qk in q])
    elif form == "gauss":
        Iq = np.array([gauss_quad_1d(qk, n) for qk in q])
    pylab.loglog(q, Iq, label="%s, n=%d"%(form, n))
    pylab.xlabel("q (1/A)")
    pylab.ylabel("Iq (1/cm)")
    pylab.title(KERNEL.__doc__ + " I(q) circular average")
    return Iq

radius = 10.
length = 1e5
NORM, KERNEL = make_cylinder(radius=radius, length=length)
long_cyl = make_long_cylinder(radius=radius, length=length)
#NORM, KERNEL = make_sphere(radius=50.)


if __name__ == "__main__":
    Q = 0.386
    for n in (20, 76, 150, 300, 1000): #, 10000, 30000):
        print("gauss-%d"%n, gauss_quad_1d(Q, n=n))
    for k in (8, 10, 13, 16, 19):
        gridded_1d(Q, n=2**k+1)
    #print("inf cyl", 0, long_cyl(Q))
    #scipy_romberg(Q)

    plot_1d(0.386, n=2000)
    plot_1d(0.5, n=2000)
    plot_1d(0.8, n=2000)
    pylab.legend()
    pylab.figure()

    q = np.logspace(-3, 0, 400)
    I1 = long_cyl(q)
    I2 = plot_Iq(q, n=2**19+1, form="trapz")
    #plot_Iq(q, n=2**16+1, form="trapz")
    #plot_Iq(q, n=2**10+1, form="trapz")
    plot_Iq(q, n=1024, form="gauss")
    #plot_Iq(q, n=300, form="gauss")
    #plot_Iq(q, n=150, form="gauss")
    #plot_Iq(q, n=76, form="gauss")
    pylab.loglog(q, long_cyl(q), label="limit")
    pylab.legend()

    pylab.figure()
    pylab.semilogx(q, (I2 - I1)/I1)

    pylab.show()



import sys

# Fake the existence of 'sas' module
class SasDataloaderLoader:
    pass
class SasDataloader:
    loader = SasDataloaderLoader()
class Sas:
    dataloader = SasDataloader()
sas = Sas()

sys.modules['sas'] = sas
sys.modules['sas.dataloader'] = sas.dataloader
sys.modules['sas.dataloader.loader'] = sas.dataloader.loader

def register(linter):
    pass



r"""
Definition
----------

This model provides the form factor, $P(q)$, for a monodisperse hollow right
angle circular cylinder (rigid tube) where the The inside and outside of the
hollow cylinder are assumed to have the same SLD and the form factor is thus
normalized by the volume of the tube (i.e. not by the total cylinder volume).

.. math::

    P(q) = \text{scale} \left<F^2\right>/V_\text{shell} + \text{background}

where the averaging $\left<\ldots\right>$ is applied only for the 1D
calculation. If Intensity is given on an absolute scale, the scale factor here
is the volume fraction of the shell.  This differs from
the :ref:`core-shell-cylinder` in that, in that case, scale is the volume
fraction of the entire cylinder (core+shell). The application might be for a
bilayer which wraps into a hollow tube and the volume fraction of material is
all in the shell, whereas the :ref:`core-shell-cylinder` model might be used for
a cylindrical micelle where the tails in the core have a different SLD than the
headgroups (in the shell) and the volume fraction of material comes fromm the
whole cyclinder.  NOTE: the hollow_cylinder represents a tube whereas the
core_shell_cylinder includes a shell layer covering the ends (end caps) as well.


The 1D scattering intensity is calculated in the following way (Guinier, 1955)

.. math::

    P(q)           &= (\text{scale})V_\text{shell}\Delta\rho^2
            \int_0^{1}\Psi^2
            \left[q_z, R_\text{outer}(1-x^2)^{1/2},
                       R_\text{core}(1-x^2)^{1/2}\right]
            \left[\frac{\sin(qHx)}{qHx}\right]^2 dx \\
    \Psi[q,y,z]    &= \frac{1}{1-\gamma^2}
            \left[ \Lambda(qy) - \gamma^2\Lambda(qz) \right] \\
    \Lambda(a)     &= 2 J_1(a) / a \\
    \gamma         &= R_\text{core} / R_\text{outer} \\
    V_\text{shell} &= \pi \left(R_\text{outer}^2 - R_\text{core}^2 \right)L \\
    J_1(x)         &= (\sin(x)-x\cdot \cos(x)) / x^2

where *scale* is a scale factor, $H = L/2$ and $J_1$ is the 1st order
Bessel function.

**NB**: The 2nd virial coefficient of the cylinder is calculated
based on the outer radius and full length, which give an the effective radius
for structure factor $S(q)$ when $P(q) \cdot S(q)$ is applied.

In the parameters,the *radius* is $R_\text{core}$ while *thickness*
is $R_\text{outer} - R_\text{core}$.

To provide easy access to the orientation of the core-shell cylinder, we define
the axis of the cylinder using two angles $\theta$ and $\phi$
(see :ref:`cylinder model <cylinder-angle-definition>`).

References
----------

#. L A Feigin and D I Svergun, *Structure Analysis by Small-Angle X-Ray and
   Neutron Scattering*, Plenum Press, New York, (1987)
#. L. Onsager, *Ann. New York Acad. Sci.*, 51 (1949) 627-659

Authorship and Verification
----------------------------

* **Author:** NIST IGOR/DANSE **Date:** pre 2010
* **Last Modified by:** Paul Butler **Date:** September 06, 2018
   (corrected VR calculation)
* **Last Reviewed by:** Paul Butler **Date:** September 06, 2018
"""
from __future__ import division

import numpy as np
from numpy import pi, inf, sin, cos

name = "hollow_cylinder"
title = ""
description = """
P(q) = scale*<f*f>/Vol + background, where f is the scattering amplitude.
radius = the radius of core
thickness = the thickness of shell
length = the total length of the cylinder
sld = SLD of the shell
sld_solvent = SLD of the solvent
background = incoherent background
"""
category = "shape:cylinder"
# pylint: disable=bad-whitespace, line-too-long
#   ["name", "units", default, [lower, upper], "type","description"],
parameters = [
    ["radius",      "Ang",     20.0, [0, inf],    "volume",      "Cylinder core radius"],
    ["thickness",   "Ang",     10.0, [0, inf],    "volume",      "Cylinder wall thickness"],
    ["length",      "Ang",    400.0, [0, inf],    "volume",      "Cylinder total length"],
    ["sld",         "1e-6/Ang^2",  6.3, [-inf, inf], "sld",      "Cylinder sld"],
    ["sld_solvent", "1e-6/Ang^2",  1,   [-inf, inf], "sld",      "Solvent sld"],
    ["theta",       "degrees", 90,   [-360, 360], "orientation", "Cylinder axis to beam angle"],
    ["phi",         "degrees",  0,   [-360, 360], "orientation", "Rotation about beam"],
    ]
# pylint: enable=bad-whitespace, line-too-long

source = ["lib/polevl.c", "lib/sas_J1.c", "lib/gauss76.c", "hollow_cylinder.c"]
have_Fq = True
radius_effective_modes = [
    "excluded volume", "equivalent outer volume sphere",
    "outer radius", "half length",
    "half outer min dimension", "half outer max dimension",
    "half outer diagonal",
    ]

def random():
    """Return a random parameter set for the model."""
    length = 10**np.random.uniform(1, 4.7)
    outer_radius = 10**np.random.uniform(1, 4.7)
    # Use a distribution with a preference for thin shell or thin core
    # Avoid core,shell radii < 1
    thickness = np.random.beta(0.5, 0.5)*(outer_radius-2) + 1
    radius = outer_radius - thickness
    pars = dict(
        length=length,
        radius=radius,
        thickness=thickness,
    )
    return pars

# Tests for structure factor support functions. These functions duplicate
# the equivalent functions from the C code for the hollow cylinder model.
def r_eff(radius, thickness, length):
    """R_eff from excluded volume"""
    radius += thickness
    return (0.5*(0.75*radius*(2.0*radius*length
                              + (radius + length)*(pi*radius + length))
                )**(1./3.))

def shell_volume(radius, thickness, length):
    """shell volume for parameter set"""
    return pi*((radius+thickness)**2-radius**2)*length

def form_shell_ratio(radius, thickness, length):
    """form:shell ratio"""
    return (radius+thickness)**2/((radius+thickness)**2 - radius**2)

q = 0.1
# april 6 2017, rkh added a 2d unit test, assume correct!
qx = q*cos(pi/6.0)
qy = q*sin(pi/6.0)
test_pars = [
    parameters[0][2], # radius
    parameters[1][2], # thickness
    parameters[2][2], # length
]
# Parameters for unit tests
tests = [
    [{}, 0.00005, 1764.926],
    [{}, 0.1, None, None,
     r_eff(*test_pars), shell_volume(*test_pars), form_shell_ratio(*test_pars),
    ],
    [{}, 0.001, 1756.76],
    [{}, (qx, qy), 2.36885476192],
]
del qx, qy, test_pars  # not necessary to delete, but cleaner



# parallelepiped model
# Note: model title and parameter table are inserted automatically
r"""
Definition
----------

This model calculates the scattering from a rectangular solid
(:numref:`parallelepiped-image`).
If you need to apply polydispersity, see also :ref:`rectangular-prism`. For
information about polarised and magnetic scattering, see
the :ref:`magnetism` documentation.

.. _parallelepiped-image:


.. figure:: img/parallelepiped_geometry.jpg

   Parallelepiped with the corresponding definition of sides.

The three dimensions of the parallelepiped (strictly here a cuboid) may be
given in *any* size order as long as the particles are randomly oriented (i.e.
take on all possible orientations see notes on 2D below). To avoid multiple fit
solutions, especially with Monte-Carlo fit methods, it may be advisable to
restrict their ranges. There may be a number of closely similar "best fits", so
some trial and error, or fixing of some dimensions at expected values, may
help.

The form factor is normalized by the particle volume and the 1D scattering
intensity $I(q)$ is then calculated as:

.. Comment by Miguel Gonzalez:
   I am modifying the original text because I find the notation a little bit
   confusing. I think that in most textbooks/papers, the notation P(Q) is
   used for the form factor (adim, P(Q=0)=1), although F(q) seems also to
   be used. But here (as for many other models), P(q) is used to represent
   the scattering intensity (in cm-1 normally). It would be good to agree on
   a common notation.

.. math::

    I(q) = \frac{\text{scale}}{V} (\Delta\rho \cdot V)^2
           \left< P(q, \alpha, \beta) \right> + \text{background}

where the volume $V = A B C$, the contrast is defined as
$\Delta\rho = \rho_\text{p} - \rho_\text{solvent}$, $P(q, \alpha, \beta)$
is the form factor corresponding to a parallelepiped oriented
at an angle $\alpha$ (angle between the long axis C and $\vec q$), and $\beta$
(the angle between the projection of the particle in the $xy$ detector plane
and the $y$ axis) and the averaging $\left<\ldots\right>$ is applied over all
orientations.

Assuming $a = A/B < 1$, $b = B /B = 1$, and $c = C/B > 1$, the
form factor is given by (Mittelbach and Porod, 1961 [#Mittelbach1961]_)

.. math::

    P(q, \alpha) = \int_0^1 \phi_Q\left(\mu \sqrt{1-\sigma^2},a\right)
        \left[S(\mu c \sigma/2)\right]^2 d\sigma

with

.. math::

    \phi_Q(\mu,a) &= \int_0^1
        \left\{S\left[\frac{\mu}{2}\cos\left(\frac{\pi}{2}u\right)\right]
               S\left[\frac{\mu a}{2}\sin\left(\frac{\pi}{2}u\right)\right]
               \right\}^2 du \\
    S(x) &= \frac{\sin x}{x} \\
    \mu &= qB

where substitution of $\sigma = cos\alpha$ and $\beta = \pi/2 \ u$ have been
applied.

For **oriented** particles, the 2D scattering intensity, $I(q_x, q_y)$, is
given as:

.. math::

    I(q_x, q_y) = \frac{\text{scale}}{V} (\Delta\rho \cdot V)^2 P(q_x, q_y)
            + \text{background}

.. Comment by Miguel Gonzalez:
   This reflects the logic of the code, as in parallelepiped.c the call
   to _pkernel returns $P(q_x, q_y)$ and then this is multiplied by
   $V^2 * (\Delta \rho)^2$. And finally outside parallelepiped.c it will be
   multiplied by scale, normalized by $V$ and the background added. But
   mathematically it makes more sense to write
   $I(q_x, q_y) = \text{scale} V \Delta\rho^2 P(q_x, q_y) + \text{background}$,
   with scale being the volume fraction.

Where $P(q_x, q_y)$ for a given orientation of the form factor is calculated as

.. math::

    P(q_x, q_y) = \left[\frac{\sin(\tfrac{1}{2}qA\cos\alpha)}{(\tfrac{1}
                   {2}qA\cos\alpha)}\right]^2
                  \left[\frac{\sin(\tfrac{1}{2}qB\cos\beta)}{(\tfrac{1}
                   {2}qB\cos\beta)}\right]^2
                  \left[\frac{\sin(\tfrac{1}{2}qC\cos\gamma)}{(\tfrac{1}
                   {2}qC\cos\gamma)}\right]^2

with

.. math::

    \cos\alpha &= \hat A \cdot \hat q, \\
    \cos\beta  &= \hat B \cdot \hat q, \\
    \cos\gamma &= \hat C \cdot \hat q


FITTING NOTES
~~~~~~~~~~~~~

#. The 2nd virial coefficient of the parallelepiped is calculated based on
   the averaged effective radius, after appropriately sorting the three
   dimensions, to give an oblate or prolate particle, $(=\sqrt{AB/\pi})$ and
   length $(= C)$ values, and used as the effective radius for
   $S(q)$ when $P(q) \cdot S(q)$ is applied.

#. For 2d data the orientation of the particle is required, described using
   angles $\theta$, $\phi$ and $\Psi$ as in the diagrams below, where $\theta$
   and $\phi$ define the orientation of the director in the laboratory reference
   frame of the beam direction ($z$) and detector plane ($x-y$ plane), while
   the angle $\Psi$ is effectively the rotational angle around the particle
   $C$ axis. For $\theta = 0$ and $\phi = 0$, $\Psi = 0$ corresponds to the
   $B$ axis oriented parallel to the y-axis of the detector with $A$ along
   the x-axis. For other $\theta$, $\phi$ values, the order of rotations
   matters. In particular, the parallelepiped must first be rotated $\theta$
   degrees in the $x-z$ plane before rotating $\phi$ degrees around the $z$
   axis (in the $x-y$ plane). Applying orientational distribution to the
   particle orientation (i.e  `jitter` to one or more of these angles) can get
   more confusing as `jitter` is defined **NOT** with respect to the laboratory
   frame but the particle reference frame. It is thus highly recommended to
   read :ref:`orientation` for further details of the calculation and angular
   dispersions.

.. note:: For 2d, constraints must be applied during fitting to ensure that the
   order of sides chosen is not altered, and hence that the correct definition
   of angles is preserved. For the default choice shown here, that means
   ensuring that the inequality $A < B < C$ is not violated,  The calculation
   will not report an error, but the results may be not correct.

.. _parallelepiped-orientation:

.. figure:: img/parallelepiped_angle_definition.png

    Definition of the angles for oriented parallelepiped, shown with $A<B<C$.

.. figure:: img/parallelepiped_angle_projection.png

    Examples of the angles for an oriented parallelepiped against the
    detector plane.

.. Comment by Paul Butler
   I am commenting this section out as we are trying to minimize the amount of
   oritentational detail here and encourage the user to go to the full
   orientation documentation so that changes can be made in just one place.
   below is the commented paragraph:
   On introducing "Orientational Distribution" in the angles, "distribution of
   theta" and "distribution of phi" parameters will appear. These are actually
   rotations about axes $\delta_1$ and $\delta_2$ of the parallelepiped,
   perpendicular to the $a$ x $c$ and $b$ x $c$ faces. (When $\theta = \phi = 0$
   these are parallel to the $Y$ and $X$ axes of the instrument.) The third
   orientation distribution, in $\psi$, is about the $c$ axis of the particle,
   perpendicular to the $a$ x $b$ face. Some experimentation may be required to
   understand the 2d patterns fully as discussed in :ref:`orientation` .


Validation
----------

Validation of the code was done by comparing the output of the 1D calculation
to the angular average of the output of a 2D calculation over all possible
angles.

References
----------

See also Nayuk [#Nayuk2012]_ and Onsager [#Onsager1949]_.

.. [#Mittelbach1961] P Mittelbach and G Porod, *Acta Physica Austriaca*,
   14 (1961) 185-211
.. [#Nayuk2012] R Nayuk and K Huber, *Z. Phys. Chem.*, 226 (2012) 837-854
.. [#Onsager1949] L. Onsager, *Ann. New York Acad. Sci.*, 51 (1949) 627-659

Authorship and Verification
----------------------------

* **Author:** NIST IGOR/DANSE **Date:** pre 2010
* **Last Modified by:**  Paul Kienzle **Date:** April 05, 2017
* **Last Reviewed by:**  Miguel Gonzales and Paul Butler **Date:** May 24,
  2018 - documentation updated
"""

import numpy as np
from numpy import inf

name = "parallelepiped"
title = "Rectangular parallelepiped with uniform scattering length density."
description = """
    I(q)= scale*V*(sld - sld_solvent)^2*P(q,alpha)+background
        P(q,alpha) = integral from 0 to 1 of ...
           phi(mu*sqrt(1-sigma^2),a) * S(mu*c*sigma/2)^2 * dsigma
        with
            phi(mu,a) = integral from 0 to 1 of ..
            (S((mu/2)*cos(pi*u/2))*S((mu*a/2)*sin(pi*u/2)))^2 * du
            S(x) = sin(x)/x
            mu = q*B
        V: Volume of the rectangular parallelepiped
        alpha: angle between the long axis of the
            parallelepiped and the q-vector for 1D
"""
category = "shape:parallelepiped"

#             ["name", "units", default, [lower, upper], "type","description"],
parameters = [["sld", "1e-6/Ang^2", 4, [-inf, inf], "sld",
               "Parallelepiped scattering length density"],
              ["sld_solvent", "1e-6/Ang^2", 1, [-inf, inf], "sld",
               "Solvent scattering length density"],
              ["length_a", "Ang", 35, [0, inf], "volume",
               "Shorter side of the parallelepiped"],
              ["length_b", "Ang", 75, [0, inf], "volume",
               "Second side of the parallelepiped"],
              ["length_c", "Ang", 400, [0, inf], "volume",
               "Larger side of the parallelepiped"],
              ["theta", "degrees", 60, [-360, 360], "orientation",
               "c axis to beam angle"],
              ["phi", "degrees", 60, [-360, 360], "orientation",
               "rotation about beam"],
              ["psi", "degrees", 60, [-360, 360], "orientation",
               "rotation about c axis"],
             ]

source = ["lib/gauss76.c", "parallelepiped.c"]
have_Fq = True
radius_effective_modes = [
    "equivalent cylinder excluded volume", "equivalent volume sphere",
    "half length_a", "half length_b", "half length_c",
    "equivalent circular cross-section", "half ab diagonal", "half diagonal",
    ]

def random():
    """Return a random parameter set for the model."""
    length = 10**np.random.uniform(1, 4.7, size=3)
    pars = dict(
        length_a=length[0],
        length_b=length[1],
        length_c=length[2],
    )
    return pars

# rkh 7/4/17 add random unit test for 2d, note make all params different,
# 2d values not tested against other codes or models
qx, qy = 0.2 * np.cos(np.pi/6.), 0.2 * np.sin(np.pi/6.)
tests = [[{}, 0.2, 0.17758004974],
         [{}, [0.2], [0.17758004974]],
         [{'theta':10.0, 'phi':20.0}, (qx, qy), 0.0089517140475],
         [{'theta':10.0, 'phi':20.0}, [(qx, qy)], [0.0089517140475]],
        ]
del qx, qy  # not necessary to delete, but cleaner



r"""
.. warning:: This model is not functioning correctly in SasView and it
             appears it has not done so for some time. Whilst the
             problem is investigated, a workaround for Case 0 below
             (the most common use case) is to use the binary_blend
             model available on the `Model Maketplace 
             <https://marketplace.sasview.org/models/124/>`_ . For further
             information, please email help@sasview.org . *The
             SasView Developers. February 2022.*

Definition
----------

Calculates the macroscopic scattering intensity for a multi-component
homogeneous mixture of polymers using the Random Phase Approximation.
This general formalism contains 10 specific cases

Case 0: C/D binary mixture of homopolymers

Case 1: C-D diblock copolymer

Case 2: B/C/D ternary mixture of homopolymers

Case 3: C/C-D mixture of a homopolymer B and a diblock copolymer C-D

Case 4: B-C-D triblock copolymer

Case 5: A/B/C/D quaternary mixture of homopolymers

Case 6: A/B/C-D mixture of two homopolymers A/B and a diblock C-D

Case 7: A/B-C-D mixture of a homopolymer A and a triblock B-C-D

Case 8: A-B/C-D mixture of two diblock copolymers A-B and C-D

Case 9: A-B-C-D tetra-block copolymer

.. note::
    These case numbers are different from those in the NIST SANS package!

The models are based on the papers by Akcasu *et al.* [1] and by
Hammouda [2] assuming the polymer follows Gaussian statistics such
that $R_g^2 = n b^2/6$ where $b$ is the statistical segment length and $n$ is
the number of statistical segment lengths. A nice tutorial on how these are
constructed and implemented can be found in chapters 28, 31 and 34, and Part H,
of Hammouda's 'SANS Toolbox' [3].

In brief, the macroscopic cross sections are derived from the general forms
for homopolymer scattering and the multiblock cross-terms while the inter,
polymer cross terms are described in the usual way by the $\chi$ parameter.

USAGE NOTES:

* Only one case can be used at any one time.
* The RPA (mean field) formalism only applies only when the multicomponent
  polymer mixture is in the homogeneous mixed-phase region.
* **Component D is assumed to be the "background" component (ie, all contrasts
  are calculated with respect to component D).** So the scattering contrast
  for a C/D blend $\rho_{C/D} = [\rho_C - \rho_D]$\ :sup:`2`.
* Depending on which case is being used, the number of fitting parameters can
  vary.

  .. Note::
    * In general the degrees of polymerization, the volume
      fractions, the molar volumes, and the neutron scattering lengths for each
      component are obtained from other methods and held fixed while The *scale*
      parameter should be held equal to unity.
    * The variables are normally the segment lengths ($b_a$, $b_b$,
      etc.) and $\chi$ parameters ($K_{ab}$, $K_{ac}$, etc).

References
----------

#. A Z Akcasu, R Klein and B Hammouda, *Macromolecules*, 26 (1993) 4136
#. B. Hammouda, *Advances in Polymer Science* 106 (1993) 87
#. B. Hammouda, *SANS Toolbox*
   https://www.ncnr.nist.gov/staff/hammouda/the_sans_toolbox.pdf.

Authorship and Verification
----------------------------

* **Author:** Boualem Hammouda - NIST IGOR/DANSE **Date:** pre 2010
* **Converted to sasmodels by:** Paul Kienzle **Date:** July 18, 2016
* **Last Modified by:** Paul Butler **Date:** March 12, 2017
* **Last Reviewed by:** Steve King **Date:** March 27, 2019
"""

from numpy import inf

name = "rpa"
title = "Random Phase Approximation"
description = """
This formalism applies to multicomponent polymer mixtures in the
homogeneous (mixed) phase region only.
Case 0: C/D binary mixture of homopolymers
Case 1: C-D diblock copolymer
Case 2: B/C/D ternary mixture of homopolymers
Case 3: B/C-D mixture of homopolymer b and diblock copolymer C-D
Case 4: B-C-D triblock copolymer
Case 5: A/B/C/D quaternary mixture of homopolymers
Case 6: A/B/C-D mixture of two homopolymers A/B and a diblock C-D
Case 7: A/B-C-D mixture of a homopolymer A and a triblock B-C-D
Case 8: A-B/C-D mixture of two diblock copolymers A-B and C-D
Case 9: A-B-C-D four-block copolymer
See details in the model function help
"""
category = "shape-independent"

CASES = [
    "C+D binary mixture",
    "C:D diblock copolymer",
    "B+C+D ternary mixture",
    "B+C:D binary mixture",
    "B:C:D triblock copolymer",
    "A+B+C+D quaternary mixture",
    "A+B+C:D ternary mixture",
    "A+B:C:D binary mixture",
    "A:B+C:D binary mixture",
    "A:B:C:D quadblock copolymer",
]

#   ["name", "units", default, [lower, upper], "type","description"],
parameters = [
    ["case_num", "", 1, [CASES], "", "Component organization"],

    ["N[4]", "", 1000.0, [1, inf], "", "Degree of polymerization"],
    ["Phi[4]", "", 0.25, [0, 1], "", "volume fraction"],
    ["v[4]", "mL/mol", 100.0, [0, inf], "", "molar volume"],
    ["L[4]", "fm", 10.0, [-inf, inf], "", "scattering length"],
    ["b[4]", "Ang", 5.0, [0, inf], "", "segment length"],

    ["K12", "", -0.0004, [-inf, inf], "", "A:B interaction parameter"],
    ["K13", "", -0.0004, [-inf, inf], "", "A:C interaction parameter"],
    ["K14", "", -0.0004, [-inf, inf], "", "A:D interaction parameter"],
    ["K23", "", -0.0004, [-inf, inf], "", "B:C interaction parameter"],
    ["K24", "", -0.0004, [-inf, inf], "", "B:D interaction parameter"],
    ["K34", "", -0.0004, [-inf, inf], "", "C:D interaction parameter"],
]


source = ["rpa.c"]
single = False

control = "case_num"
HIDE_ALL = set("Phi4".split())
HIDE_A = set("N1 Phi1 v1 L1 b1 K12 K13 K14".split()).union(HIDE_ALL)
HIDE_AB = set("N2 Phi2 v2 L2 b2 K23 K24".split()).union(HIDE_A)
def hidden(case_num):
    """
    Return a list of parameters to hide depending on the multiplicity parameter.
    """
    case_num = int(case_num+0.5)
    if case_num < 2:
        return HIDE_AB
    elif case_num < 5:
        return HIDE_A
    else:
        return HIDE_ALL

# TODO: no random parameters generated for RPA
